# Copyright 2022 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for `von_mises.py`."""
import functools

from absl.testing import absltest
from absl.testing import parameterized

import chex
from distrax._src.distributions import von_mises
from distrax._src.utils import equivalence
import jax
import jax.numpy as jnp
import numpy as np
from scipy import stats as sp_stats
import scipy.special
from tensorflow_probability.substrates import jax as tfp

jax.config.update('jax_threefry_partitionable', False)


class VonMisesTest(equivalence.EquivalenceTest):

  def setUp(self):
    super().setUp()
    self._init_distr_cls(von_mises.VonMises)
    self.loc = np.reshape(np.linspace(-5., 5., 7), [-1, 1])
    self.concentration = np.reshape(np.logspace(-3., 3., 7), [1, -1])
    self.rng = np.random.default_rng(317070)

  @parameterized.named_parameters(
      ('1d std vonmises', (0, 1)),
      ('2d std vonmises', (np.zeros(2), np.ones(2))),
      ('rank 2 std vonmises', (np.zeros((3, 2)), np.ones((3, 2)))),
      ('broadcasted loc', (0, np.ones(3))),
      ('broadcasted concentration', (np.ones(3), 1)),
  )
  def test_event_shape(self, distr_params):
    super()._test_event_shape(distr_params, dict())

  @parameterized.named_parameters(
      ('0d concentration', ()),
      ('1d concentration', (4,)),
      ('2d concentration', (3, 4)),
  )
  def test_loc_shape_properties(self, shape):
    loc = self.rng.uniform()
    concentration = self.rng.uniform(size=shape)
    dist = self.distrax_cls(loc=loc, concentration=concentration)
    self.assertion_fn(rtol=1e-3)(dist.loc, loc)
    self.assertion_fn(rtol=1e-3)(dist.concentration, concentration)
    self.assertEqual(dist.event_shape, ())
    self.assertEqual(dist.batch_shape, shape)

  @parameterized.named_parameters(
      ('0d loc', ()),
      ('1d loc', (4,)),
      ('2d loc', (3, 4)),
  )
  def test_concentration_shape_properties(self, shape):
    loc = self.rng.uniform(size=shape)
    concentration = self.rng.uniform()
    dist = self.distrax_cls(loc=loc, concentration=concentration)
    self.assertion_fn(rtol=1e-3)(dist.loc, loc)
    self.assertion_fn(rtol=1e-3)(dist.concentration, concentration)
    self.assertEqual(dist.event_shape, ())
    self.assertEqual(dist.batch_shape, shape)

  @chex.all_variants
  @parameterized.named_parameters(
      ('1d std vonmises, no shape', (0, 1), ()),
      ('1d std vonmises, int shape', (0, 1), 1),
      ('1d std vonmises, 1-tuple shape', (0, 1), (1,)),
      ('1d std vonmises, 2-tuple shape', (0, 1), (2, 2)),
      ('2d std vonmises, no shape', (np.zeros(2), np.ones(2)), ()),
      ('2d std vonmises, int shape', ([0, 0], [1, 1]), 1),
      ('2d std vonmises, 1-tuple shape', (np.zeros(2), np.ones(2)), (1,)),
      ('2d std vonmises, 2-tuple shape', ([0, 0], [1, 1]), (2, 2)),
      ('rank 2 std vonmises, 2-tuple shape', (np.zeros((3, 2)), np.ones(
          (3, 2))), (2, 2)),
      ('broadcasted loc', (0, np.ones(3)), (2, 2)),
      ('broadcasted scale', (np.ones(3), 1), ()),
  )
  def test_sample_shape(self, distr_params, sample_shape):
    distr_params = (np.asarray(distr_params[0], dtype=np.float32),
                    np.asarray(distr_params[1], dtype=np.float32))
    super()._test_sample_shape(distr_params, dict(), sample_shape)

  @chex.all_variants
  @parameterized.named_parameters(
      ('1d std vonmises, no shape', (0, 1), ()),
      ('1d std vonmises, int shape', (0, 1), 1),
      ('1d std vonmises, 1-tuple shape', (0, 1), (1,)),
      ('1d std vonmises, 2-tuple shape', (0, 1), (2, 2)),
      ('2d std vonmises, no shape', (np.zeros(2), np.ones(2)), ()),
      ('2d std vonmises, int shape', ([0, 0], [1, 1]), 1),
      ('2d std vonmises, 1-tuple shape', (np.zeros(2), np.ones(2)), (1,)),
      ('2d std vonmises, 2-tuple shape', ([0, 0], [1, 1]), (2, 2)),
      ('rank 2 std vonmises, 2-tuple shape', (np.zeros((3, 2)), np.ones(
          (3, 2))), (2, 2)),
      ('broadcasted loc', (0, np.ones(3)), (2, 2)),
      ('broadcasted scale', (np.ones(3), 1), ()),
  )
  def test_sample_and_log_prob(self, distr_params, sample_shape):
    distr_params = (np.asarray(distr_params[0], dtype=np.float32),
                    np.asarray(distr_params[1], dtype=np.float32))
    super()._test_sample_and_log_prob(
        dist_args=distr_params,
        dist_kwargs=dict(),
        sample_shape=sample_shape,
        assertion_fn=self.assertion_fn(rtol=1e-2))

  @chex.all_variants
  @parameterized.named_parameters(
      ('sample, float16', 'sample', jnp.float16),
      ('sample, float32', 'sample', jnp.float32),
      ('sample_and_log_prob, float16', 'sample_and_log_prob', jnp.float16),
      ('sample_and_log_prob, float32', 'sample_and_log_prob', jnp.float32),
  )
  def test_sample_dtype(self, method, dtype):
    dist = self.distrax_cls(
        loc=self.loc.astype(dtype),
        concentration=self.concentration.astype(dtype),
    )
    samples = self.variant(getattr(dist, method))(seed=self.key)
    samples = samples[0] if method == 'sample_and_log_prob' else samples
    self.assertEqual(samples.dtype, dist.dtype)
    self.assertEqual(samples.dtype, dtype)

  @chex.all_variants
  @parameterized.named_parameters(
      ('1d dist, 1d value', (0, 1), 1),
      ('1d dist, 2d value', (0., 1.), np.array([1, 2])),
      ('2d dist, 1d value', (np.zeros(2), np.ones(2)), 1),
      ('2d broadcasted dist, 1d value', (np.zeros(2), 1), 1),
      ('2d dist, 2d value', (np.zeros(2), np.ones(2)), np.array([1, 2])),
      ('1d dist, 1d value, edge case', (0, 1), np.pi),
  )
  def test_method_with_input(self, distr_params, value):
    distr_params = (np.asarray(distr_params[0], dtype=np.float32),
                    np.asarray(distr_params[1], dtype=np.float32))
    value = np.asarray(value, dtype=np.float32)
    for method in [
        'log_prob', 'prob', 'cdf', 'log_cdf', 'survival_function',
        'log_survival_function'
    ]:
      with self.subTest(method):
        super()._test_attribute(
            attribute_string=method,
            dist_args=distr_params,
            dist_kwargs={},
            call_args=(value,),
            assertion_fn=self.assertion_fn(rtol=1e-2))

  @chex.all_variants(with_pmap=False)
  @parameterized.named_parameters(
      ('entropy', (0., 1.), 'entropy'),
      ('mean', (0, 1), 'mean'),
      ('mean from 1d params', ([-1, 1], [1, 2]), 'mean'),
      ('variance', (0, 1), 'variance'),
      ('variance from np params', (np.ones(2), np.ones(2)), 'variance'),
      ('stddev', (0, 1), 'stddev'),
      ('stddev from rank 2 params', (np.ones((2, 3)), np.ones(
          (2, 3))), 'stddev'),
      ('mode', (0, 1), 'mode'),
  )
  def test_method(self, distr_params, function_string):
    distr_params = (np.asarray(distr_params[0], dtype=np.float32),
                    np.asarray(distr_params[1], dtype=np.float32))
    super()._test_attribute(
        function_string,
        distr_params,
        assertion_fn=self.assertion_fn(rtol=1e-2))

  @chex.all_variants(with_pmap=False)
  @parameterized.named_parameters(
      ('kl distrax_to_distrax', 'kl_divergence', 'distrax_to_distrax'),
      ('kl distrax_to_tfp', 'kl_divergence', 'distrax_to_tfp'),
      ('kl tfp_to_distrax', 'kl_divergence', 'tfp_to_distrax'),
      ('cross-ent distrax_to_distrax', 'cross_entropy', 'distrax_to_distrax'),
      ('cross-ent distrax_to_tfp', 'cross_entropy', 'distrax_to_tfp'),
      ('cross-ent tfp_to_distrax', 'cross_entropy', 'tfp_to_distrax'))
  def test_with_two_distributions(self, function_string, mode_string):
    super()._test_with_two_distributions(
        attribute_string=function_string,
        mode_string=mode_string,
        dist1_kwargs={
            'loc': self.rng.standard_normal((4, 1, 2)),
            'concentration': np.asarray([[0.8, 0.2], [0.1, 1.2], [1.4, 3.1]]),
        },
        dist2_kwargs={
            'loc': self.rng.standard_normal((3, 2)),
            'concentration': 0.1 + self.rng.standard_normal((4, 1, 2)),
        },
        assertion_fn=self.assertion_fn(rtol=1e-2))

  def test_jittable(self):
    super()._test_jittable(
        (np.zeros((3,)), np.ones((3,))),
        assertion_fn=functools.partial(
            np.testing.assert_allclose, rtol=1e-04, atol=1e-04
        )
    )

  @parameterized.named_parameters(
      ('single element', 2),
      ('range', slice(-1)),
      ('range_2', (slice(None), slice(-1))),
      ('ellipsis', (Ellipsis, -1)),
  )
  def test_slice(self, slice_):
    loc = jnp.array(self.rng.standard_normal((3, 4, 5)))
    concentration = jnp.array(self.rng.standard_normal((3, 4, 5)))
    dist = self.distrax_cls(loc=loc, concentration=concentration)
    self.assertion_fn(rtol=1e-2)(dist[slice_].mean(), loc[slice_])

  def test_slice_different_parameterization(self):
    loc = jnp.array(self.rng.standard_normal((4)))
    concentration = jnp.array(self.rng.standard_normal((3, 4)))
    dist = self.distrax_cls(loc=loc, concentration=concentration)
    self.assertion_fn(rtol=1e-2)(dist[0].mean(), loc)  # Not slicing loc.
    self.assertion_fn(rtol=1e-2)(dist[0].concentration, concentration[0])

  def test_von_mises_log_pdf(self):
    locs_v = .1
    concentrations_v = .2
    x = np.array([2., 3., 4., 5., 6., 7.])
    vm = self.distrax_cls(locs_v, concentrations_v)
    expected_log_prob = sp_stats.vonmises.logpdf(  # pytype: disable=module-attr
        x,
        concentrations_v,
        loc=locs_v
    )
    log_prob = vm.log_prob(x)
    np.testing.assert_allclose(
        expected_log_prob, log_prob, rtol=1e-04, atol=1e-04
    )

  def test_von_mises_log_pdf_uniform(self):
    x = np.array([2., 3., 4., 5., 6., 7.])
    vm = self.distrax_cls(.1, 0.)
    log_prob = vm.log_prob(x)
    expected_log_prob = np.array([-np.log(2. * np.pi)] * 6)
    np.testing.assert_allclose(
        expected_log_prob, log_prob, rtol=1e-04, atol=1e-04
    )

  def test_von_mises_pdf(self):
    locs_v = .1
    concentrations_v = .2
    x = np.array([2., 3., 4., 5., 6., 7.])
    vm = self.distrax_cls(locs_v, concentrations_v)
    prob = vm.prob(x)
    expected_prob = sp_stats.vonmises.pdf(  # pytype: disable=module-attr
        x, concentrations_v, loc=locs_v
    )
    np.testing.assert_allclose(expected_prob, prob, rtol=1e-04, atol=1e-04)

  def test_von_mises_pdf_uniform(self):
    x = np.array([2., 3., 4., 5., 6., 7.])
    vm = self.distrax_cls(1., 0.)
    prob = vm.prob(x)
    expected_prob = np.array([1. / (2. * np.pi)] * 6)
    np.testing.assert_allclose(expected_prob, prob, rtol=1.5e-7, atol=1e-7)

  def test_von_mises_cdf(self):
    # We follow the scipy definition for cdf when loc=0 and x is in [-pi, pi].
    locs_v = np.zeros(shape=(7, 1, 1))
    concentrations_v = np.reshape(np.logspace(-3., 3., 7), [1, -1, 1])
    x = np.reshape(np.linspace(-np.pi, np.pi, 7), [1, 1, -1])
    vm = self.distrax_cls(locs_v, concentrations_v)
    cdf = vm.cdf(x)
    expected_cdf = sp_stats.vonmises.cdf(  # pytype: disable=module-attr
        x, concentrations_v, loc=locs_v
    )
    np.testing.assert_allclose(expected_cdf, cdf, atol=1e-4, rtol=1e-4)

  def test_von_mises_survival_function(self):
    locs_v = np.reshape(np.linspace(-5, 5, 7), [-1, 1, 1])
    concentrations_v = np.reshape(np.logspace(-3., 3., 7), [1, -1, 1])
    x = np.reshape(np.linspace(-5, 5, 7), [1, 1, -1])
    vm = self.distrax_cls(locs_v, concentrations_v)
    cdf = vm.cdf(x)
    surv = vm.survival_function(x)
    np.testing.assert_allclose(surv, 1 - cdf, atol=1e-4, rtol=1e-4)

  def test_von_mises_cdf_out_of_bounds(self):
    locs_v = np.reshape(np.linspace(-np.pi, np.pi, 7), [-1, 1, 1])
    concentrations_v = np.reshape(np.logspace(-3., 3., 7), [1, -1, 1])
    vm = self.distrax_cls(locs_v, concentrations_v)
    x = np.linspace(-5 * np.pi, -np.pi, 7)
    cdf = vm.cdf(x)
    expected_cdf = 0.
    np.testing.assert_allclose(expected_cdf, cdf, rtol=1.5e-7, atol=1e-7)

    x = np.linspace(np.pi, 5 * np.pi, 7)
    cdf = vm.cdf(x)
    expected_cdf = 1.
    np.testing.assert_allclose(expected_cdf, cdf, rtol=1.5e-7, atol=1e-7)

  def test_von_mises_log_cdf(self):
    locs_v = np.reshape(np.linspace(-5, 5, 7), [-1, 1, 1])
    concentrations_v = np.reshape(np.logspace(-3., 3., 7), [1, -1, 1])
    x = np.reshape(np.linspace(-5, 5, 7), [1, 1, -1])
    vm = self.distrax_cls(locs_v, concentrations_v)
    cdf = vm.cdf(x)
    logcdf = vm.log_cdf(x)
    np.testing.assert_allclose(np.log(cdf), logcdf, atol=1e-4, rtol=1e-4)

  def test_von_mises_log_survival(self):
    locs_v = np.reshape(np.linspace(-5, 5, 7), [-1, 1, 1])
    concentrations_v = np.reshape(np.logspace(-3., 3., 7), [1, -1, 1])
    x = np.reshape(np.linspace(-5, 5, 7), [1, 1, -1])
    vm = self.distrax_cls(locs_v, concentrations_v)
    surv = vm.survival_function(x)
    logsurv = vm.log_survival_function(x)
    np.testing.assert_allclose(np.log(surv), logsurv, atol=1e-4, rtol=1e-4)

  def test_von_mises_cdf_uniform(self):
    x = np.linspace(-np.pi, np.pi, 7)
    vm = self.distrax_cls(0., 0.)
    cdf = vm.cdf(x)
    expected_cdf = (x + np.pi) / (2. * np.pi)
    np.testing.assert_allclose(expected_cdf, cdf, rtol=1.5e-7, atol=1e-7)

  def test_von_mises_cdf_gradient_simple(self):
    n = 10
    locs = jnp.array([1.] * n)
    concentrations = np.logspace(-3, 3, n)
    x = np.linspace(-5, 5, n)
    def f(x, l, c):
      vm = self.distrax_cls(l, c)
      cdf = vm.cdf(x)
      return cdf
    jax.test_util.check_grads(f, (x, locs, concentrations), order=1)

  def test_von_mises_sample_gradient(self):
    n = 10
    locs = jnp.array([1.] * n)
    concentrations = np.logspace(-3, 3, n)
    def f(l, c):
      vm = self.distrax_cls(l, c)
      x = vm.sample(seed=1)
      return x
    jax.test_util.check_grads(
        f,
        (locs, concentrations),
        order=1,
        rtol=0.1
    )

  def test_von_mises_uniform_sample_gradient(self):
    def f(c):
      vm = self.distrax_cls(0., c)
      x = vm.sample(seed=1)
      return x
    # The finite difference is not very accurate, but the analytic gradient
    # should not be zero.
    self.assertNotEqual(jax.grad(f)(0.), 0)

  @parameterized.named_parameters(
      ('small concentration', 1.),
      ('medium concentration', 10.),
      ('large concentration', 1000.),
      ('uniform', 1e-6),
  )
  def test_von_mises_sample_gradient_comparison(self, concentration):
    # Compares the von Mises sampling gradient against the reference
    # implementation from tensorflow_probability.
    locs = 0.
    def f(seed, l, c):
      vm = self.distrax_cls(l, c)
      x = vm.sample(seed=seed)  # pylint: disable=cell-var-from-loop
      return x
    jax_sample_and_grad = jax.value_and_grad(f, argnums=2)

    def samples_grad(s, concentration):
      broadcast_concentration = concentration
      _, dcdf_dconcentration = tfp.math.value_and_gradient(
          lambda conc: tfp.distributions.von_mises.von_mises_cdf(s, conc),
          broadcast_concentration)
      inv_prob = np.exp(-concentration * (np.cos(s) - 1.)) * (
          (2. * np.pi) * scipy.special.i0e(concentration)
      )
      # Computes the implicit derivative,
      # dz = dconc * -(dF(z; conc) / dconc) / p(z; conc)
      dsamples = -dcdf_dconcentration * inv_prob
      return dsamples

    for seed in range(10):
      sample, sample_grad = jax_sample_and_grad(
          seed, jnp.array(locs), jnp.array(concentration)
      )
      comparison = samples_grad(sample, concentration)
      np.testing.assert_allclose(
          comparison, sample_grad, rtol=1e-06, atol=1e-06
      )

  def test_von_mises_sample_moments(self):
    locs_v = np.array([-1., 0.3, 2.3])
    concentrations_v = np.array([1., 2., 10.])
    vm = self.distrax_cls(locs_v, concentrations_v)

    n = 1000
    samples = vm.sample(sample_shape=(n,), seed=1)

    expected_mean = vm.mean()
    actual_mean = jnp.arctan2(
        jnp.mean(jnp.sin(samples), axis=0),
        jnp.mean(jnp.cos(samples), axis=0),
    )

    expected_variance = vm.variance()
    standardized_samples = samples - vm.mean()
    variance_samples = jnp.mean(1. - jnp.cos(standardized_samples), axis=0)

    np.testing.assert_allclose(actual_mean, expected_mean, rtol=0.1)
    np.testing.assert_allclose(
        variance_samples, expected_variance, rtol=0.1
    )

  def test_von_mises_sample_variance_uniform(self):
    vm = self.distrax_cls(1., 0.)

    n = 1000
    samples = vm.sample(sample_shape=(n,), seed=1)

    # For circular uniform distribution, the mean is not well-defined,
    # so only checking the variance.
    expected_variance = 1.
    standardized_samples = samples - vm.mean()
    variance_samples = jnp.mean(1. - jnp.cos(standardized_samples), axis=0)

    np.testing.assert_allclose(
        variance_samples, expected_variance, rtol=0.1
    )

  def test_von_mises_sample_extreme_concentration(self):
    loc = jnp.array([1., np.nan, 1., 1., np.nan])
    min_value = np.finfo(np.float32).min
    max_value = np.finfo(np.float32).max
    concentration = jnp.array([min_value, 1., max_value, np.nan, np.nan])
    vm = self.distrax_cls(loc, concentration)

    samples = vm.sample(seed=1)
    # Check that it does not end up in an infinite loop.
    self.assertEqual(samples.shape, (5,))

  def test_von_mises_sample_ks_test(self):
    concentrations_v = np.logspace(-3, 3, 7)
    # We are fixing the location to zero. The reason is that for loc != 0,
    # scipy's von Mises distribution CDF becomes shifted, so it's no longer
    # in [0, 1], but is in something like [-0.3, 0.7]. This breaks kstest.
    vm = self.distrax_cls(0., concentrations_v)
    n = 1000
    sample_values = vm.sample(sample_shape=(n,), seed=1)
    self.assertEqual(sample_values.shape, (n, 7))

    fails = 0
    trials = 0
    for concentrationi, concentration in enumerate(concentrations_v):
      s = sample_values[:, concentrationi]
      trials += 1
      p = sp_stats.kstest(
          s,
          sp_stats.vonmises(concentration).cdf  # pytype: disable=not-callable
      )[1]
      if p <= 0.05:
        fails += 1
    self.assertLess(fails, trials * 0.1)

  def test_von_mises_sample_uniform_ks_test(self):
    locs_v = np.linspace(-10., 10., 7)
    vm = self.distrax_cls(locs_v, 0.)

    n = 1000
    sample_values = vm.sample(sample_shape=(n,), seed=1)
    self.assertEqual(sample_values.shape, (n, 7))

    fails = 0
    trials = 0
    for loci, _ in enumerate(locs_v):
      s = sample_values[:, loci]
      # [-pi, pi] -> [0, 1]
      s = (s + np.pi) / (2. * np.pi)
      trials += 1
      # Compare to the CDF of Uniform(0, 1) random variable.
      p = sp_stats.kstest(s, sp_stats.uniform.cdf)[1]
      if p <= 0.05:
        fails += 1
    self.assertLess(fails, trials * 0.1)

  def test_von_mises_sample_average_gradient(self):
    loc = jnp.array([1.] * 7)
    concentration = np.logspace(-3, 3, 7)
    grad_ys = np.ones(7, dtype=np.float32)
    n = 1000

    def loss(loc, concentration):
      vm = self.distrax_cls(loc, concentration)
      samples = vm.sample(sample_shape=(n,), seed=1)
      return jnp.mean(samples, axis=0)

    grad_loc, grad_concentration = jnp.vectorize(
        jax.grad(loss, argnums=(0, 1)),
        signature='(),()->(),()',
    )(loc, concentration)

    # dsamples / dloc = 1 => dloss / dloc = dloss / dsamples = grad_ys
    np.testing.assert_allclose(grad_loc, grad_ys, atol=1e-1, rtol=1e-1)
    np.testing.assert_allclose(grad_concentration, [0.]*7, atol=1e-1, rtol=1e-1)

  def test_von_mises_sample_circular_variance_gradient(self):
    loc = jnp.array([1.] * 7)
    concentration = np.logspace(-3, 3, 7)
    n = 1000

    def loss(loc, concentration):
      vm = self.distrax_cls(loc, concentration)
      samples = vm.sample(sample_shape=(n,), seed=1)
      return jnp.mean(1-jnp.cos(samples-loc), axis=0)

    grad_loc, grad_concentration = jnp.vectorize(
        jax.grad(loss, argnums=(0, 1)),
        signature='(),()->(),()',
    )(loc, concentration)

    def analytical_loss(concentration):
      i1e = jax.scipy.special.i1e(concentration)
      i0e = jax.scipy.special.i0e(concentration)
      return 1. - i1e / i0e

    expected_grad_concentration = jnp.vectorize(
        jax.grad(analytical_loss)
    )(concentration)

    np.testing.assert_allclose(grad_loc, [0.] * 7, atol=1e-2, rtol=1e-2)
    np.testing.assert_allclose(
        grad_concentration, expected_grad_concentration, atol=1e-1, rtol=1e-1)


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Distribution representing a Bijector applied to a Distribution."""

from typing import Optional, Tuple, Union

from distrax._src.bijectors import bijector as bjct_base
from distrax._src.distributions import distribution as dist_base
from distrax._src.utils import conversion
import jax
import jax.numpy as jnp
from tensorflow_probability.substrates import jax as tfp

tfd = tfp.distributions

PRNGKey = dist_base.PRNGKey
Array = dist_base.Array
DistributionLike = dist_base.DistributionLike
BijectorLike = bjct_base.BijectorLike
EventT = dist_base.EventT


class Transformed(dist_base.Distribution):
  """Distribution of a random variable transformed by a bijective function.

  Let `X` be a continuous random variable and `Y = f(X)` be a random variable
  transformed by a differentiable bijection `f` (a "bijector"). Given the
  distribution of `X` (the "base distribution") and the bijector `f`, this class
  implements the distribution of `Y` (also known as the pushforward of the base
  distribution through `f`).

  The probability density of `Y` can be computed by:

  `log p(y) = log p(x) - log|det J(f)(x)|`

  where `p(x)` is the probability density of `X` (the "base density") and
  `J(f)(x)` is the Jacobian matrix of `f`, both evaluated at `x = f^{-1}(y)`.

  Sampling from a Transformed distribution involves two steps: sampling from the
  base distribution `x ~ p(x)` and then evaluating `y = f(x)`. The first step
  is agnostic to the possible batch dimensions of the bijector `f(x)`. For
  example:
  ```
    dist = distrax.Normal(loc=0., scale=1.)
    bij = distrax.ScalarAffine(shift=jnp.asarray([3., 3., 3.]))
    transformed_dist = distrax.Transformed(distribution=dist, bijector=bij)
    samples = transformed_dist.sample(seed=0, sample_shape=())
    print(samples)  # [2.7941577, 2.7941577, 2.7941577]
  ```

  Note: the `batch_shape`, `event_shape`, and `dtype` properties of the
  transformed distribution, as well as the `kl_divergence` method, are computed
  on-demand via JAX tracing when requested. This assumes that the `forward`
  function of the bijector is traceable; that is, it is a pure function that
  does not contain run-time branching. Functions that do not strictly meet this
  requirement can still be used, but we cannot guarantee that the shapes, dtype,
  and KL computations involving the transformed distribution can be correctly
  obtained.
  """

  equiv_tfp_cls = tfd.TransformedDistribution

  def __init__(self, distribution: DistributionLike, bijector: BijectorLike):
    """Initializes a Transformed distribution.

    Args:
      distribution: the base distribution. Can be either a Distrax distribution
        or a TFP distribution.
      bijector: a differentiable bijective transformation. Can be a Distrax
        bijector, a TFP bijector, or a callable to be wrapped by `Lambda`.
    """
    super().__init__()
    distribution = conversion.as_distribution(distribution)
    bijector = conversion.as_bijector(bijector)

    event_shape = distribution.event_shape
    # Check if event shape is a tuple of integers (i.e. not nested).
    if not (isinstance(event_shape, tuple) and
            all(isinstance(i, int) for i in event_shape)):
      raise ValueError(
          f"'Transformed' currently only supports distributions with Array "
          f"events (i.e. not nested). Received '{distribution.name}' with "
          f"event shape '{distribution.event_shape}'.")

    if len(event_shape) != bijector.event_ndims_in:
      raise ValueError(
          f"Base distribution '{distribution.name}' has event shape "
          f"{distribution.event_shape}, but bijector '{bijector.name}' expects "
          f"events to have {bijector.event_ndims_in} dimensions. Perhaps use "
          f"`distrax.Block` or `distrax.Independent`?")

    self._distribution = distribution
    self._bijector = bijector
    self._batch_shape = None
    self._event_shape = None
    self._dtype = None

  @property
  def distribution(self):
    """The base distribution."""
    return self._distribution

  @property
  def bijector(self):
    """The bijector representing the transformation."""
    return self._bijector

  def _infer_shapes_and_dtype(self):
    """Infer the batch shape, event shape, and dtype by tracing `forward`."""
    dummy_shape = self.distribution.batch_shape + self.distribution.event_shape
    dummy = jnp.zeros(dummy_shape, dtype=self.distribution.dtype)
    shape_dtype = jax.eval_shape(self.bijector.forward, dummy)

    self._dtype = shape_dtype.dtype

    if self.bijector.event_ndims_out == 0:
      self._event_shape = ()
      self._batch_shape = shape_dtype.shape
    else:
      # pylint: disable-next=invalid-unary-operand-type
      self._event_shape = shape_dtype.shape[-self.bijector.event_ndims_out:]
      # pylint: disable-next=invalid-unary-operand-type
      self._batch_shape = shape_dtype.shape[:-self.bijector.event_ndims_out]

  @property
  def dtype(self) -> jnp.dtype:
    """See `Distribution.dtype`."""
    if self._dtype is None:
      self._infer_shapes_and_dtype()
    assert self._dtype is not None  # By _infer_shapes_and_dtype()
    return self._dtype

  @property
  def event_shape(self) -> Tuple[int, ...]:
    """See `Distribution.event_shape`."""
    if self._event_shape is None:
      self._infer_shapes_and_dtype()
    assert self._event_shape is not None  # By _infer_shapes_and_dtype()
    return self._event_shape

  @property
  def batch_shape(self) -> Tuple[int, ...]:
    """See `Distribution.batch_shape`."""
    if self._batch_shape is None:
      self._infer_shapes_and_dtype()
    assert self._batch_shape is not None  # By _infer_shapes_and_dtype()
    return self._batch_shape

  def log_prob(self, value: EventT) -> Array:
    """See `Distribution.log_prob`."""
    x, ildj_y = self.bijector.inverse_and_log_det(value)
    lp_x = self.distribution.log_prob(x)
    lp_y = lp_x + ildj_y
    return lp_y

  def _sample_n(self, key: PRNGKey, n: int) -> Array:
    """Returns `n` samples."""
    x = self.distribution.sample(seed=key, sample_shape=n)
    y = jax.vmap(self.bijector.forward)(x)
    return y

  def _sample_n_and_log_prob(self, key: PRNGKey, n: int) -> Tuple[Array, Array]:
    """Returns `n` samples and their log probs.

    This function is more efficient than calling `sample` and `log_prob`
    separately, because it uses only the forward methods of the bijector. It
    also works for bijectors that don't implement inverse methods.

    Args:
      key: PRNG key.
      n: Number of samples to generate.

    Returns:
      A tuple of `n` samples and their log probs.
    """
    x, lp_x = self.distribution.sample_and_log_prob(seed=key, sample_shape=n)
    y, fldj = jax.vmap(self.bijector.forward_and_log_det)(x)
    lp_y = jax.vmap(jnp.subtract)(lp_x, fldj)
    return y, lp_y

  def mean(self) -> Array:
    """Calculates the mean."""
    if self.bijector.is_constant_jacobian:
      return self.bijector.forward(self.distribution.mean())
    else:
      raise NotImplementedError(
          "`mean` is not implemented for this transformed distribution, "
          "because its bijector's Jacobian is not known to be constant.")

  def mode(self) -> Array:
    """Calculates the mode."""
    if self.bijector.is_constant_log_det:
      return self.bijector.forward(self.distribution.mode())
    else:
      raise NotImplementedError(
          "`mode` is not implemented for this transformed distribution, "
          "because its bijector's Jacobian determinant is not known to be "
          "constant.")

  def entropy(  # pylint: disable=arguments-differ
      self,
      input_hint: Optional[Array] = None) -> Array:
    """Calculates the Shannon entropy (in Nats).

    Only works for bijectors with constant Jacobian determinant.

    Args:
      input_hint: an example sample from the base distribution, used to compute
        the constant forward log-determinant. If not specified, it is computed
        using a zero array of the shape and dtype of a sample from the base
        distribution.

    Returns:
      the entropy of the distribution.

    Raises:
      NotImplementedError: if bijector's Jacobian determinant is not known to be
                           constant.
    """
    if self.bijector.is_constant_log_det:
      if input_hint is None:
        shape = self.distribution.batch_shape + self.distribution.event_shape
        input_hint = jnp.zeros(shape, dtype=self.distribution.dtype)
      entropy = self.distribution.entropy()
      fldj = self.bijector.forward_log_det_jacobian(input_hint)
      return entropy + fldj
    else:
      raise NotImplementedError(
          "`entropy` is not implemented for this transformed distribution, "
          "because its bijector's Jacobian determinant is not known to be "
          "constant.")


def _kl_divergence_transformed_transformed(
    dist1: Union[Transformed, tfd.TransformedDistribution],
    dist2: Union[Transformed, tfd.TransformedDistribution],
    *unused_args,
    input_hint: Optional[Array] = None,
    **unused_kwargs,
) -> Array:
  """Obtains the KL divergence between two Transformed distributions.

  This computes the KL divergence between two Transformed distributions with the
  same bijector. If the two Transformed distributions do not have the same
  bijector, an error is raised. To determine if the bijectors are equal, this
  method proceeds as follows:
  - If both bijectors are the same instance of a Distrax bijector, then they are
    declared equal.
  - If not the same instance, we check if they are equal according to their
    `same_as` predicate.
  - Otherwise, the string representation of the Jaxpr of the `forward` method
    of each bijector is compared. If both string representations are equal, the
    bijectors are declared equal.
  - Otherwise, the bijectors cannot be guaranteed to be equal and an error is
    raised.

  Args:
    dist1: A Transformed distribution.
    dist2: A Transformed distribution.
    input_hint: an example sample from the base distribution, used to trace the
      `forward` method. If not specified, it is computed using a zero array of
      the shape and dtype of a sample from the base distribution.

  Returns:
    Batchwise `KL(dist1 || dist2)`.

  Raises:
    NotImplementedError: If bijectors are not known to be equal.
    ValueError: If the base distributions do not have the same `event_shape`.
  """
  if dist1.distribution.event_shape != dist2.distribution.event_shape:
    raise ValueError(
        f"The two base distributions do not have the same event shape: "
        f"{dist1.distribution.event_shape} and "
        f"{dist2.distribution.event_shape}.")

  bij1 = conversion.as_bijector(dist1.bijector)  # conversion needed for TFP
  bij2 = conversion.as_bijector(dist2.bijector)

  # Check if the bijectors are different.
  if bij1 != bij2 and not bij1.same_as(bij2):
    if input_hint is None:
      input_hint = jnp.zeros(
          dist1.distribution.event_shape, dtype=dist1.distribution.dtype)
    jaxpr_bij1 = jax.make_jaxpr(bij1.forward)(input_hint).jaxpr
    jaxpr_bij2 = jax.make_jaxpr(bij2.forward)(input_hint).jaxpr
    if str(jaxpr_bij1) != str(jaxpr_bij2):
      raise NotImplementedError(
          f"The KL divergence cannot be obtained because it is not possible to "
          f"guarantee that the bijectors {dist1.bijector.name} and "
          f"{dist2.bijector.name} of the Transformed distributions are "
          f"equal. If possible, use the same instance of a Distrax bijector.")

  return dist1.distribution.kl_divergence(dist2.distribution)


# Register the KL functions with TFP.
tfd.RegisterKL(Transformed, Transformed)(_kl_divergence_transformed_transformed)
tfd.RegisterKL(Transformed.equiv_tfp_cls, Transformed)(
    _kl_divergence_transformed_transformed)
tfd.RegisterKL(Transformed, Transformed.equiv_tfp_cls)(
    _kl_divergence_transformed_transformed)



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Linear bijector."""

import abc
from typing import Sequence, Tuple

from distrax._src.bijectors import bijector as base
import jax.numpy as jnp

Array = base.Array


class Linear(base.Bijector, metaclass=abc.ABCMeta):
  """Base class for linear bijectors.

  This class provides a base class for bijectors defined as `f(x) = Ax`,
  where `A` is a `DxD` matrix and `x` is a `D`-dimensional vector.
  """

  def __init__(self,
               event_dims: int,
               batch_shape: Sequence[int],
               dtype: jnp.dtype):
    """Initializes a `Linear` bijector.

    Args:
      event_dims: the dimensionality `D` of the event `x`. It is assumed that
        `x` is a vector of length `event_dims`.
      batch_shape: the batch shape of the bijector.
      dtype: the data type of matrix `A`.
    """
    super().__init__(event_ndims_in=1, is_constant_jacobian=True)
    self._event_dims = event_dims
    self._batch_shape = tuple(batch_shape)
    self._dtype = dtype

  @property
  def matrix(self) -> Array:
    """The matrix `A` of the transformation.

    To be optionally implemented in a subclass.

    Returns:
      An array of shape `batch_shape + (event_dims, event_dims)` and data type
      `dtype`.
    """
    raise NotImplementedError(
        f"Linear bijector {self.name} does not implement `matrix`.")

  @property
  def event_dims(self) -> int:
    """The dimensionality `D` of the event `x`."""
    return self._event_dims

  @property
  def batch_shape(self) -> Tuple[int, ...]:
    """The batch shape of the bijector."""
    return self._batch_shape

  @property
  def dtype(self) -> jnp.dtype:
    """The data type of matrix `A`."""
    return self._dtype



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Dirichlet distribution."""

from typing import Tuple, Union

import chex
from distrax._src.distributions import distribution
from distrax._src.distributions.beta import Beta
from distrax._src.utils import conversion
from distrax._src.utils import math
import jax
import jax.numpy as jnp
from tensorflow_probability.substrates import jax as tfp

tfd = tfp.distributions

Array = chex.Array
Numeric = chex.Numeric
PRNGKey = chex.PRNGKey
EventT = distribution.EventT


class Dirichlet(distribution.Distribution):
  """Dirichlet distribution with concentration parameter `alpha`.

  The PDF of a Dirichlet distributed random variable `X`, where `X` lives in the
  simplex `(0, 1)^K` with `sum_{k=1}^{K} X_k = 1`, is given by,
  ```
  p(x; alpha) = ( prod_{k=1}^{K} x_k ** (alpha_k - 1) ) / B(alpha)
  ```
  where `B(alpha)` is the multivariate beta function, and the concentration
  parameters `alpha_k > 0`.

  Note that the support of the distribution does not include `x_k = 0` nor
  `x_k = 1`.
  """

  equiv_tfp_cls = tfd.Dirichlet

  def __init__(self, concentration: Array):
    """Initializes a Dirichlet distribution.

    Args:
      concentration: Concentration parameter `alpha` of the distribution. It
        must be an array of length `K >= 2` containing positive values
        (additional dimensions index batches).
    """
    super().__init__()
    self._concentration = conversion.as_float_array(concentration)
    if self._concentration.ndim < 1:
      raise ValueError(
          'The concentration parameter must have at least one dimension.')
    if self._concentration.shape[-1] < 2:
      raise ValueError(
          'The last dimension of the concentration parameter must be '
          'at least 2.')
    self._log_normalization_constant = math.log_beta_multivariate(
        self._concentration)

  @property
  def event_shape(self) -> Tuple[int, ...]:
    """Shape of event of distribution samples."""
    return self._concentration.shape[-1:]

  @property
  def batch_shape(self) -> Tuple[int, ...]:
    """Shape of batch of distribution samples."""
    return self._concentration.shape[:-1]

  @property
  def concentration(self) -> Array:
    """Concentration parameter `alpha` of the distribution."""
    return self._concentration

  def _sample_n(self, key: PRNGKey, n: int) -> Array:
    """See `Distribution._sample_n`."""
    out_shape = (n,) + self.batch_shape
    dtype = self._concentration.dtype
    rnd = jax.random.dirichlet(
        key, alpha=self._concentration, shape=out_shape, dtype=dtype)
    return rnd

  def log_prob(self, value: EventT) -> Array:
    """See `Distribution.log_prob`."""
    return (jnp.sum((self._concentration - 1.) * jnp.log(value), axis=-1)
            - self._log_normalization_constant)

  def entropy(self) -> Array:
    """Calculates the Shannon entropy (in nats)."""
    sum_concentration = jnp.sum(self._concentration, axis=-1)
    return (
        self._log_normalization_constant
        + ((sum_concentration - self._concentration.shape[-1])
           * jax.lax.digamma(sum_concentration))
        - jnp.sum((self._concentration - 1.) *
                  jax.lax.digamma(self._concentration), axis=-1)
    )

  def mean(self) -> Array:
    """Calculates the mean."""
    return self._concentration / jnp.sum(
        self._concentration, axis=-1, keepdims=True)

  def mode(self) -> Array:
    """Calculates the mode.

    Returns:
      The mode, an array of shape `batch_shape + event_shape`. If any
      `alpha_k <= 1`, the returned value is `jnp.nan`.
    """
    result_if_valid = (self._concentration - 1.) / jnp.sum(
        self._concentration - 1., axis=-1, keepdims=True)
    return jnp.where(
        jnp.all(self._concentration > 1., axis=-1, keepdims=True),
        result_if_valid,
        jnp.nan)

  def variance(self) -> Array:
    """Calculates the variance."""
    sum_concentration = jnp.sum(self._concentration, axis=-1, keepdims=True)
    norm_concentration = self._concentration / sum_concentration
    return norm_concentration * (1. - norm_concentration) / (
        sum_concentration + 1.)

  def covariance(self) -> Array:
    """Calculates the covariance.

    Returns:
      An array of shape `batch_shape + event_shape + event_shape` with the
      covariance of the distribution.
    """
    sum_concentration = jnp.sum(self._concentration, axis=-1, keepdims=True)
    norm_concentration = self._concentration / sum_concentration
    norm_over_sum = norm_concentration / (sum_concentration + 1.)
    cov = - jnp.expand_dims(norm_over_sum, axis=-1) * jnp.expand_dims(
        norm_concentration, axis=-2)
    cov += jnp.vectorize(jnp.diag, signature='(k)->(k,k)')(norm_over_sum)
    return cov

  def __getitem__(self, index) -> 'Dirichlet':
    """See `Distribution.__getitem__`."""
    index = distribution.to_batch_shape_index(self.batch_shape, index)
    return Dirichlet(concentration=self.concentration[index])


DirichletLike = Union[Beta, tfd.Beta, Dirichlet, tfd.Dirichlet]


def _obtain_concentration(dist: DirichletLike) -> Array:
  """Returns the concentration parameters of the input distribution."""
  if isinstance(dist, Dirichlet):
    concentration = dist.concentration
  elif isinstance(dist, Beta):
    concentration = jnp.stack((dist.alpha, dist.beta), axis=-1)
  elif isinstance(dist, tfd.Beta):
    concentration = jnp.stack(
        (dist.concentration1, dist.concentration0), axis=-1)
  elif isinstance(dist, tfd.Dirichlet):
    concentration = dist.concentration
  else:
    raise ValueError(f'Unsupported distribution: {type(dist)}.')
  return concentration


def _kl_divergence_dirichlet_dirichlet(
    dist1: DirichletLike,
    dist2: DirichletLike,
    *unused_args,
    **unused_kwargs,
) -> Array:
  """KL divergence KL(dist1 || dist2) between two Dirichlet distributions.

  Args:
    dist1: A Dirichlet or Beta distribution.
    dist2: A Dirichlet or Beta distribution.

  Returns:
    Batchwise `KL(dist1 || dist2)`.
  """
  concentration1 = _obtain_concentration(dist1)
  concentration2 = _obtain_concentration(dist2)
  if concentration1.shape[-1] != concentration2.shape[-1]:
    raise ValueError(
        f'The two distributions must have the same event dimension, but got '
        f'{concentration1.shape[-1]} and {concentration2.shape[-1]} '
        f'dimensions.')
  sum_concentration1 = jnp.sum(concentration1, axis=-1, keepdims=True)
  t1 = (math.log_beta_multivariate(concentration2)
        - math.log_beta_multivariate(concentration1))
  t2 = jnp.sum((concentration1 - concentration2) * (
      jax.lax.digamma(concentration1) - jax.lax.digamma(sum_concentration1)),
               axis=-1)
  return t1 + t2


# Register the KL functions with TFP.
tfd.RegisterKL(Dirichlet, Dirichlet)(_kl_divergence_dirichlet_dirichlet)
tfd.RegisterKL(Dirichlet, Dirichlet.equiv_tfp_cls)(
    _kl_divergence_dirichlet_dirichlet)
tfd.RegisterKL(Dirichlet.equiv_tfp_cls, Dirichlet)(
    _kl_divergence_dirichlet_dirichlet)

tfd.RegisterKL(Dirichlet, Beta)(_kl_divergence_dirichlet_dirichlet)
tfd.RegisterKL(Beta, Dirichlet)(_kl_divergence_dirichlet_dirichlet)
tfd.RegisterKL(Dirichlet, tfd.Beta)(_kl_divergence_dirichlet_dirichlet)
tfd.RegisterKL(tfd.Beta, Dirichlet)(_kl_divergence_dirichlet_dirichlet)
tfd.RegisterKL(tfd.Dirichlet, Beta)(_kl_divergence_dirichlet_dirichlet)
tfd.RegisterKL(Beta, tfd.Dirichlet)(_kl_divergence_dirichlet_dirichlet)



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for `shift.py`."""

from absl.testing import absltest
from absl.testing import parameterized

import chex
from distrax._src.bijectors.shift import Shift
from distrax._src.bijectors.tanh import Tanh
import jax
import jax.numpy as jnp
import numpy as np


class ShiftTest(parameterized.TestCase):

  def test_jacobian_is_constant_property(self):
    bijector = Shift(jnp.ones((4,)))
    self.assertTrue(bijector.is_constant_jacobian)
    self.assertTrue(bijector.is_constant_log_det)

  def test_properties(self):
    bijector = Shift(jnp.array([1., 2., 3.]))
    np.testing.assert_array_equal(bijector.shift, np.array([1., 2., 3.]))

  @chex.all_variants
  @parameterized.parameters(
      {'batch_shape': (), 'param_shape': ()},
      {'batch_shape': (3,), 'param_shape': ()},
      {'batch_shape': (), 'param_shape': (3,)},
      {'batch_shape': (2, 3), 'param_shape': (2, 3)},
  )
  def test_forward_methods(self, batch_shape, param_shape):
    bijector = Shift(jnp.ones(param_shape))
    prng = jax.random.PRNGKey(42)
    x = jax.random.normal(prng, batch_shape)
    output_shape = jnp.broadcast_shapes(batch_shape, param_shape)
    y1 = self.variant(bijector.forward)(x)
    logdet1 = self.variant(bijector.forward_log_det_jacobian)(x)
    y2, logdet2 = self.variant(bijector.forward_and_log_det)(x)
    self.assertEqual(y1.shape, output_shape)
    self.assertEqual(y2.shape, output_shape)
    self.assertEqual(logdet1.shape, output_shape)
    self.assertEqual(logdet2.shape, output_shape)
    np.testing.assert_allclose(y1, x + 1., 1e-6)
    np.testing.assert_allclose(y2, x + 1., 1e-6)
    np.testing.assert_allclose(logdet1, 0., 1e-6)
    np.testing.assert_allclose(logdet2, 0., 1e-6)

  @chex.all_variants
  @parameterized.parameters(
      {'batch_shape': (), 'param_shape': ()},
      {'batch_shape': (3,), 'param_shape': ()},
      {'batch_shape': (), 'param_shape': (3,)},
      {'batch_shape': (2, 3), 'param_shape': (2, 3)},
  )
  def test_inverse_methods(self, batch_shape, param_shape):
    bijector = Shift(jnp.ones(param_shape))
    prng = jax.random.PRNGKey(42)
    y = jax.random.normal(prng, batch_shape)
    output_shape = jnp.broadcast_shapes(batch_shape, param_shape)
    x1 = self.variant(bijector.inverse)(y)
    logdet1 = self.variant(bijector.inverse_log_det_jacobian)(y)
    x2, logdet2 = self.variant(bijector.inverse_and_log_det)(y)
    self.assertEqual(x1.shape, output_shape)
    self.assertEqual(x2.shape, output_shape)
    self.assertEqual(logdet1.shape, output_shape)
    self.assertEqual(logdet2.shape, output_shape)
    np.testing.assert_allclose(x1, y - 1., 1e-6)
    np.testing.assert_allclose(x2, y - 1., 1e-6)
    np.testing.assert_allclose(logdet1, 0., 1e-6)
    np.testing.assert_allclose(logdet2, 0., 1e-6)

  def test_jittable(self):
    @jax.jit
    def f(x, b):
      return b.forward(x)

    bij = Shift(jnp.ones((4,)))
    x = np.zeros((4,))
    f(x, bij)

  def test_same_as_itself(self):
    bij = Shift(jnp.ones((4,)))
    self.assertTrue(bij.same_as(bij))

  def test_not_same_as_others(self):
    bij = Shift(jnp.ones((4,)))
    other = Shift(jnp.zeros((4,)))
    self.assertFalse(bij.same_as(other))
    self.assertFalse(bij.same_as(Tanh()))


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Unconstrained affine bijector."""

from typing import Tuple

from distrax._src.bijectors import bijector as base
import jax
import jax.numpy as jnp

Array = base.Array


def check_affine_parameters(matrix: Array, bias: Array) -> None:
  """Checks that `matrix` and `bias` have valid shapes.

  Args:
    matrix: a matrix, or a batch of matrices.
    bias: a vector, or a batch of vectors.

  Raises:
    ValueError: if the shapes of `matrix` and `bias` are invalid.
  """
  if matrix.ndim < 2:
    raise ValueError(f"`matrix` must have at least 2 dimensions, got"
                     f" {matrix.ndim}.")
  if bias.ndim < 1:
    raise ValueError("`bias` must have at least 1 dimension.")
  if matrix.shape[-2] != matrix.shape[-1]:
    raise ValueError(f"`matrix` must be square; instead, it has shape"
                     f" {matrix.shape[-2:]}.")
  if matrix.shape[-1] != bias.shape[-1]:
    raise ValueError(f"`matrix` and `bias` have inconsistent shapes: `matrix`"
                     f" is {matrix.shape[-2:]}, `bias` is {bias.shape[-1:]}.")


class UnconstrainedAffine(base.Bijector):
  """An unconstrained affine bijection.

  This bijector is a linear-plus-bias transformation `f(x) = Ax + b`, where `A`
  is a `D x D` square matrix and `b` is a `D`-dimensional vector.

  The bijector is invertible if and only if `A` is an invertible matrix. It is
  the responsibility of the user to make sure that this is the case; the class
  will make no attempt to verify that the bijector is invertible.

  The Jacobian determinant is equal to `det(A)`. The inverse is computed by
  solving the linear system `Ax = y - b`.

  WARNING: Both the determinant and the inverse cost `O(D^3)` to compute. Thus,
  this bijector is recommended only for small `D`.
  """

  def __init__(self, matrix: Array, bias: Array):
    """Initializes an `UnconstrainedAffine` bijector.

    Args:
      matrix: the matrix `A` in `Ax + b`. Must be square and invertible. Can
        also be a batch of matrices.
      bias: the vector `b` in `Ax + b`. Can also be a batch of vectors.
    """
    check_affine_parameters(matrix, bias)
    super().__init__(event_ndims_in=1, is_constant_jacobian=True)
    self._batch_shape = jnp.broadcast_shapes(matrix.shape[:-2], bias.shape[:-1])
    self._matrix = matrix
    self._bias = bias
    self._logdet = jnp.linalg.slogdet(matrix)[1]

  @property
  def matrix(self) -> Array:
    """The matrix `A` of the transformation."""
    return self._matrix

  @property
  def bias(self) -> Array:
    """The shift `b` of the transformation."""
    return self._bias

  def forward(self, x: Array) -> Array:
    """Computes y = f(x)."""
    self._check_forward_input_shape(x)

    def unbatched(single_x, matrix, bias):
      return matrix @ single_x + bias

    batched = jnp.vectorize(unbatched, signature="(m),(m,m),(m)->(m)")
    return batched(x, self._matrix, self._bias)

  def forward_log_det_jacobian(self, x: Array) -> Array:
    """Computes log|det J(f)(x)|."""
    self._check_forward_input_shape(x)
    batch_shape = jax.lax.broadcast_shapes(self._batch_shape, x.shape[:-1])
    return jnp.broadcast_to(self._logdet, batch_shape)

  def forward_and_log_det(self, x: Array) -> Tuple[Array, Array]:
    """Computes y = f(x) and log|det J(f)(x)|."""
    return self.forward(x), self.forward_log_det_jacobian(x)

  def inverse(self, y: Array) -> Array:
    """Computes x = f^{-1}(y)."""
    self._check_inverse_input_shape(y)

    def unbatched(single_y, matrix, bias):
      return jnp.linalg.solve(matrix, single_y - bias)

    batched = jnp.vectorize(unbatched, signature="(m),(m,m),(m)->(m)")
    return batched(y, self._matrix, self._bias)

  def inverse_log_det_jacobian(self, y: Array) -> Array:
    """Computes log|det J(f^{-1})(y)|."""
    return -self.forward_log_det_jacobian(y)

  def inverse_and_log_det(self, y: Array) -> Tuple[Array, Array]:
    """Computes x = f^{-1}(y) and log|det J(f^{-1})(y)|."""
    return self.inverse(y), self.inverse_log_det_jacobian(y)

  def same_as(self, other: base.Bijector) -> bool:
    """Returns True if this bijector is guaranteed to be the same as `other`."""
    if type(other) is UnconstrainedAffine:  # pylint: disable=unidiomatic-typecheck
      return all((
          self.matrix is other.matrix,
          self.bias is other.bias,
      ))
    return False



# Copyright 2022 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for `mixture_of_two.py`."""

from absl.testing import absltest
from absl.testing import parameterized

import chex
from distrax._src.distributions import mixture_of_two
from distrax._src.utils import equivalence
import haiku as hk
import jax
import jax.numpy as jnp
import numpy as np
from tensorflow_probability.substrates import jax as tfp


BATCH_SIZE = 5
PROPORTION = 0.3


class MixtureOfTwoTest(equivalence.EquivalenceTest):

  def setUp(self):
    super().setUp()
    self._init_distr_cls(mixture_of_two.MixtureOfTwo)
    components = self._get_components(hk.PRNGSequence(0))
    self.component_a = components[0]
    self.component_b = components[1]
    self.tfp_mixture = components[2]

  def assertion_fn(self, rtol: float = 1e-3):
    return lambda x, y: np.testing.assert_allclose(x, y, rtol=rtol)

  def _get_components(self, rng_seq):
    loc = jax.random.normal(next(rng_seq), (BATCH_SIZE, 2))
    scale = jax.nn.sigmoid(jax.random.normal(next(rng_seq), (BATCH_SIZE, 2)))
    component_a = tfp.distributions.Normal(loc=loc[:, 0], scale=scale[:, 0])
    component_b = tfp.distributions.Normal(loc=loc[:, 1], scale=scale[:, 1])
    full_proportion = jnp.full([BATCH_SIZE], PROPORTION)
    tfp_mixture = tfp.distributions.MixtureSameFamily(
        tfp.distributions.Categorical(probs=jnp.stack(
            [full_proportion, 1 - full_proportion], axis=-1)),
        components_distribution=tfp.distributions.Normal(loc, scale))
    return component_a, component_b, tfp_mixture

  def test_mixture_methods(self):
    rng_seq = hk.PRNGSequence(0)
    mix = self.distrax_cls(PROPORTION, self.component_a, self.component_b)

    sample_shape = (8, 1024)
    sample = self.tfp_mixture.sample(
        sample_shape=sample_shape, seed=next(rng_seq))
    other_sample = mix.sample(sample_shape=sample_shape, seed=next(rng_seq))
    chex.assert_equal_shape([sample, other_sample])
    np.testing.assert_allclose(
        sample.mean(axis=[0, 1]), other_sample.mean(axis=[0, 1]), atol=1e-1)
    np.testing.assert_allclose(
        sample.std(axis=[0, 1]), other_sample.std(axis=[0, 1]), atol=1e-1)
    np.testing.assert_allclose(
        self.tfp_mixture.log_prob(sample),
        mix.log_prob(sample), atol=1e-3)

  @chex.all_variants(with_pmap=False)
  @parameterized.named_parameters(
      ('empty shape', ()),
      ('int shape', 10),
      ('2-tuple shape', (10, 20)),
  )
  def test_sample_and_log_prob(self, sample_shape):
    mix = self.distrax_cls(PROPORTION, self.component_a, self.component_b)
    expected_event = mix.sample(
        seed=jax.random.PRNGKey(42), sample_shape=sample_shape)
    expected_log_prob = mix.log_prob(expected_event)
    event, log_prob = self.variant(
        mix.sample_and_log_prob, static_argnames='sample_shape')(
            seed=jax.random.PRNGKey(42), sample_shape=sample_shape)

    np.testing.assert_allclose(expected_log_prob, log_prob, atol=1e-3)
    np.testing.assert_allclose(expected_event, event, atol=1e-3)

  def test_jitable(self):
    @jax.jit
    def jitted_function(event, dist):
      return dist.log_prob(event)

    mix = self.distrax_cls(PROPORTION, self.component_a, self.component_b)
    event = mix.sample(seed=jax.random.PRNGKey(4242))
    log_prob = mix.log_prob(event)
    jitted_log_prob = jitted_function(event, mix)

    chex.assert_trees_all_close(
        jitted_log_prob, log_prob, atol=1e-4, rtol=1e-4)

  def test_prob_a(self):
    mix = self.distrax_cls(PROPORTION, self.component_a, self.component_b)
    self.assertEqual(mix.prob_a, PROPORTION)

  def test_prob_b(self):
    mix = self.distrax_cls(PROPORTION, self.component_a, self.component_b)
    self.assertEqual(mix.prob_b, 1. - PROPORTION)

  def test_batch_shape(self):
    mix = self.distrax_cls(PROPORTION, self.component_a, self.component_b)
    self.assertEqual(mix.batch_shape, (BATCH_SIZE,))
    self.assertEqual(mix.batch_shape, (BATCH_SIZE,))

  def test_event_shape(self):
    mix = self.distrax_cls(PROPORTION, self.component_a, self.component_b)
    self.assertEqual(mix.event_shape, ())
    self.assertEqual(mix.event_shape, ())

  @parameterized.named_parameters(
      ('single element', 1, ()),
      ('range', slice(-1), (4,)),
  )
  def test_slice(self, slice_, expected_batch_shape):
    mix = self.distrax_cls(PROPORTION, self.component_a, self.component_b)
    sliced_dist = mix[slice_]
    self.assertEqual(sliced_dist.batch_shape, expected_batch_shape)
    self.assertEqual(sliced_dist.event_shape, mix.event_shape)
    self.assertIsInstance(sliced_dist, self.distrax_cls)

  def test_invalid_parameters(self):
    rng_seq = hk.PRNGSequence(0)
    loc = jax.random.normal(next(rng_seq), (BATCH_SIZE,))
    scale = jax.nn.sigmoid(jax.random.normal(next(rng_seq), (BATCH_SIZE,)))
    concentration = jax.random.normal(next(rng_seq), (BATCH_SIZE,))
    with self.assertRaisesRegex(ValueError, 'must have the same event shape'):
      component_a = tfp.distributions.Normal(loc=loc, scale=scale)
      component_b = tfp.distributions.Dirichlet(concentration=concentration)
      self.distrax_cls(PROPORTION, component_a, component_b)

    with self.assertRaisesRegex(ValueError, 'must have the same batch shape'):
      component_a = tfp.distributions.Normal(loc=loc, scale=scale)
      component_b = tfp.distributions.Normal(loc=loc[1:], scale=scale[1:])
      self.distrax_cls(PROPORTION, component_a, component_b)

    with self.assertRaisesRegex(ValueError, 'must have the same dtype'):
      component_a = tfp.distributions.Normal(
          loc=loc.astype(jnp.float16), scale=scale.astype(jnp.float16))
      component_b = tfp.distributions.Normal(loc=loc, scale=scale)
      self.distrax_cls(PROPORTION, component_a, component_b)


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""LogStddevNormal distribution."""

import math
from typing import Optional

import chex
from distrax._src.distributions import distribution
from distrax._src.distributions import normal
from distrax._src.utils import conversion
import jax
import jax.numpy as jnp
from tensorflow_probability.substrates import jax as tfp

tfd = tfp.distributions

Array = chex.Array
Numeric = chex.Numeric


class LogStddevNormal(normal.Normal):
  """Normal distribution with `log_scale` parameter.

  The `LogStddevNormal` has three parameters: `loc`, `log_scale`, and
  (optionally) `max_scale`. The distribution is a univariate normal
  distribution with mean equal to `loc` and scale parameter (i.e., stddev) equal
  to `exp(log_scale)` if `max_scale` is None. If `max_scale` is not None, a soft
  thresholding is applied to obtain the scale parameter of the normal, so that
  its log is given by `log(max_scale) - softplus(log(max_scale) - log_scale)`.
  """

  def __init__(self,
               loc: Numeric,
               log_scale: Numeric,
               max_scale: Optional[float] = None):
    """Initializes a LogStddevNormal distribution.

    Args:
      loc: Mean of the distribution.
      log_scale: Log of the distribution's scale (before the soft thresholding
        applied when `max_scale` is not None).
      max_scale: Maximum value of the scale that this distribution will saturate
        at. This parameter can be useful for numerical stability. It is not a
        hard maximum; rather, we compute `log(scale)` as per the formula:
        `log(max_scale) - softplus(log(max_scale) - log_scale)`.
    """
    self._max_scale = max_scale
    if max_scale is not None:
      max_log_scale = math.log(max_scale)
      self._log_scale = max_log_scale - jax.nn.softplus(
          max_log_scale - conversion.as_float_array(log_scale))
    else:
      self._log_scale = conversion.as_float_array(log_scale)
    scale = jnp.exp(self._log_scale)
    super().__init__(loc, scale)

  @property
  def log_scale(self) -> Array:
    """The log standard deviation (after thresholding, if applicable)."""
    return jnp.broadcast_to(self._log_scale, self.batch_shape)

  def __getitem__(self, index) -> 'LogStddevNormal':
    """See `Distribution.__getitem__`."""
    index = distribution.to_batch_shape_index(self.batch_shape, index)
    return LogStddevNormal(
        loc=self.loc[index],
        log_scale=self.log_scale[index],
        max_scale=self._max_scale)


def _kl_logstddevnormal_logstddevnormal(
    dist1: LogStddevNormal, dist2: LogStddevNormal,
    *unused_args, **unused_kwargs) -> Array:
  """Calculates the batched KL divergence between two LogStddevNormal's.

  Args:
    dist1: A LogStddevNormal distribution.
    dist2: A LogStddevNormal distribution.

  Returns:
    Batchwise KL(dist1 || dist2).
  """
  # KL[N(u_a, s_a^2) || N(u_b, s_b^2)] between two Gaussians:
  # (s_a^2 + (u_a - u_b)^2)/(2*s_b^2) + log(s_b) - log(s_a) - 1/2.
  variance1 = jnp.square(dist1.scale)
  variance2 = jnp.square(dist2.scale)
  return ((variance1 + jnp.square(dist1.loc - dist2.loc)) / (2.0 * variance2) +
          dist2.log_scale - dist1.log_scale - 0.5)


# Register the KL function.
tfd.RegisterKL(LogStddevNormal, LogStddevNormal)(
    _kl_logstddevnormal_logstddevnormal)



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Softmax distribution."""

from typing import Any, Union

import chex
from distrax._src.distributions import categorical
from distrax._src.distributions import distribution
import jax.numpy as jnp


Array = chex.Array


class Softmax(categorical.Categorical):
  """Categorical implementing a softmax over logits, with given temperature.

  Given a set of logits, the probability mass is distributed such that each
  index `i` has probability `exp(logits[i]/τ)/Σ(exp(logits/τ)` where τ is a
  scalar `temperature` parameter such that for τ→0, the distribution
  becomes fully greedy, and for τ→∞ the distribution becomes fully uniform.
  """

  def __init__(self,
               logits: Array,
               temperature: float = 1.,
               dtype: Union[jnp.dtype, type[Any]] = int):
    """Initializes a Softmax distribution.

    Args:
      logits: Logit transform of the probability of each category.
      temperature: Softmax temperature τ.
      dtype: The type of event samples.
    """
    self._temperature = temperature
    self._unscaled_logits = logits
    scaled_logits = logits / temperature
    super().__init__(logits=scaled_logits, dtype=dtype)

  @property
  def temperature(self) -> float:
    """The softmax temperature parameter."""
    return self._temperature

  @property
  def unscaled_logits(self) -> Array:
    """The logits of the distribution before the temperature scaling."""
    return self._unscaled_logits

  def __getitem__(self, index) -> 'Softmax':
    """See `Distribution.__getitem__`."""
    index = distribution.to_batch_shape_index(self.batch_shape, index)
    return Softmax(
        logits=self.unscaled_logits[index],
        temperature=self.temperature,
        dtype=self.dtype)



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for `clipped.py`."""

from absl.testing import absltest
from absl.testing import parameterized

import chex
from distrax._src.distributions import clipped
from distrax._src.distributions import logistic
from distrax._src.distributions import normal
import jax
import jax.numpy as jnp
import numpy as np


MINIMUM = -1.0
MAXIMUM = 1.0
LOC = MINIMUM
SCALE = 0.1
SIZE = 3


class ClippedTest(parameterized.TestCase):

  @parameterized.parameters([
      [clipped.ClippedLogistic, logistic.Logistic],
      [clipped.ClippedNormal, normal.Normal],
  ])
  def test_clipped_logprob(self, factory, unclipped_factory):
    distribution = factory(
        loc=LOC, scale=SCALE, minimum=MINIMUM, maximum=MAXIMUM)
    unclipped = unclipped_factory(loc=LOC, scale=SCALE)

    np.testing.assert_allclose(
        unclipped.log_prob(0.0),
        distribution.log_prob(0.0))
    np.testing.assert_allclose(
        unclipped.log_prob(0.8),
        distribution.log_prob(0.8))

    # Testing outside of the boundary.
    self.assertEqual(-np.inf, distribution.log_prob(MINIMUM - 0.1))
    self.assertEqual(-np.inf, distribution.log_prob(MAXIMUM + 0.1))

  @parameterized.parameters([
      [clipped.ClippedLogistic, logistic.Logistic],
      [clipped.ClippedNormal, normal.Normal],
  ])
  def test_batched_clipped_logprob(self, factory, unclipped_factory):
    distribution = factory(
        loc=jnp.array([LOC]*SIZE),
        scale=jnp.array([SCALE]*SIZE),
        minimum=MINIMUM,
        maximum=MAXIMUM)
    unclipped = unclipped_factory(loc=LOC, scale=SCALE)

    np.testing.assert_allclose(
        unclipped.log_prob(jnp.array([0.0]*SIZE)),
        distribution.log_prob(jnp.array([0.0]*SIZE)))
    np.testing.assert_allclose(
        unclipped.log_prob(jnp.array([0.8]*SIZE)),
        distribution.log_prob(jnp.array([0.8]*SIZE)))

    # Testing outside of the boundary.
    np.testing.assert_allclose(
        -np.inf, distribution.log_prob(jnp.array([MINIMUM - 0.1]*SIZE)))
    np.testing.assert_allclose(
        -np.inf, distribution.log_prob(jnp.array([MAXIMUM + 0.1]*SIZE)))

  @parameterized.parameters([
      [clipped.ClippedLogistic, logistic.Logistic],
      [clipped.ClippedNormal, normal.Normal],
  ])
  def test_clipped_sampled_and_logprob(self, factory, unclipped_factory):
    distribution = factory(
        loc=LOC, scale=SCALE, minimum=MINIMUM, maximum=MAXIMUM)
    unclipped = unclipped_factory(loc=LOC, scale=SCALE)

    for rng in jax.random.split(jax.random.PRNGKey(42), 5):
      sample, log_prob = distribution.sample_and_log_prob(seed=rng)
      unclipped_sample, unclipped_log_prob = unclipped.sample_and_log_prob(
          seed=rng)
      if float(unclipped_sample) > MAXIMUM:
        np.testing.assert_allclose(sample, MAXIMUM, atol=1e-5)
      elif float(unclipped_sample) < MINIMUM:
        np.testing.assert_allclose(sample, MINIMUM, atol=1e-5)
      else:
        np.testing.assert_allclose(sample, unclipped_sample, atol=1e-5)
        np.testing.assert_allclose(log_prob, unclipped_log_prob, atol=1e-5)

  @parameterized.parameters([
      [clipped.ClippedLogistic, logistic.Logistic],
      [clipped.ClippedNormal, normal.Normal],
  ])
  def test_clipped_sample(self, factory, unclipped_factory):
    distribution = factory(
        loc=LOC, scale=SCALE, minimum=MINIMUM, maximum=MAXIMUM)
    unclipped = unclipped_factory(loc=LOC, scale=SCALE)

    for rng in jax.random.split(jax.random.PRNGKey(42), 5):
      sample = distribution.sample(seed=rng)
      unclipped_sample = unclipped.sample(seed=rng)
      if float(unclipped_sample) > MAXIMUM:
        np.testing.assert_allclose(sample, MAXIMUM, atol=1e-5)
      elif float(unclipped_sample) < MINIMUM:
        np.testing.assert_allclose(sample, MINIMUM, atol=1e-5)
      else:
        np.testing.assert_allclose(sample, unclipped_sample, atol=1e-5)

  @parameterized.parameters([
      [clipped.ClippedLogistic],
      [clipped.ClippedNormal],
  ])
  def test_extremes(self, factory):
    minimum = -1.0
    maximum = 1.0
    scale = 0.01

    # Using extreme loc.
    distribution = factory(
        loc=minimum, scale=scale, minimum=minimum, maximum=maximum)
    self.assertTrue(np.isfinite(distribution.log_prob(minimum)))
    self.assertTrue(np.isfinite(distribution.log_prob(maximum)))

    distribution = factory(
        loc=maximum, scale=scale, minimum=minimum, maximum=maximum)
    self.assertTrue(np.isfinite(distribution.log_prob(minimum)))
    self.assertTrue(np.isfinite(distribution.log_prob(maximum)))

  def test_jitable(self):
    minimum = -1.0
    maximum = 1.0
    loc = minimum
    scale = 0.1

    @jax.jit
    def jitted_function(event, dist):
      return dist.log_prob(event)

    dist = clipped.ClippedLogistic(
        loc=loc, scale=scale, minimum=minimum, maximum=maximum)
    event = dist.sample(seed=jax.random.PRNGKey(4242))
    log_prob = dist.log_prob(event)
    jitted_log_prob = jitted_function(event, dist)

    chex.assert_trees_all_close(
        jitted_log_prob, log_prob, atol=1e-4, rtol=1e-4)

  def test_properties(self):
    dist = clipped.ClippedLogistic(
        loc=LOC, scale=SCALE, minimum=MINIMUM, maximum=MAXIMUM)
    np.testing.assert_allclose(dist.minimum, MINIMUM, atol=1e-5)
    np.testing.assert_allclose(dist.maximum, MAXIMUM, atol=1e-5)
    dist = clipped.ClippedLogistic(
        loc=jnp.array([LOC]*SIZE),
        scale=jnp.array([SCALE]*SIZE),
        minimum=MINIMUM,
        maximum=MAXIMUM)
    np.testing.assert_allclose(dist.minimum, MINIMUM, atol=1e-5)
    np.testing.assert_allclose(dist.maximum, MAXIMUM, atol=1e-5)

  def test_min_max_broadcasting(self):
    dist = clipped.ClippedLogistic(
        loc=jnp.array([LOC]*SIZE),
        scale=jnp.array([SCALE]*SIZE),
        minimum=MINIMUM,
        maximum=MAXIMUM)
    self.assertEqual(dist.minimum.shape, (SIZE,))
    self.assertEqual(dist.minimum.shape, (SIZE,))

  def test_batch_shape(self):
    dist = clipped.ClippedLogistic(
        loc=jnp.array([LOC]*SIZE),
        scale=jnp.array([SCALE]*SIZE),
        minimum=MINIMUM,
        maximum=MAXIMUM)
    self.assertEqual(dist.batch_shape, (SIZE,))
    self.assertEqual(dist.batch_shape, (SIZE,))

  def test_event_shape(self):
    dist = clipped.ClippedLogistic(
        loc=jnp.array([LOC]*SIZE),
        scale=jnp.array([SCALE]*SIZE),
        minimum=jnp.array([MINIMUM]*SIZE),
        maximum=jnp.array([MAXIMUM]*SIZE))
    self.assertEqual(dist.event_shape, ())
    self.assertEqual(dist.event_shape, ())


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for `laplace.py`."""

from absl.testing import absltest
from absl.testing import parameterized

import chex
from distrax._src.distributions import laplace
from distrax._src.utils import equivalence
import jax.experimental
import jax.numpy as jnp
import numpy as np


class LaplaceTest(equivalence.EquivalenceTest):

  def setUp(self):
    super().setUp()
    self._init_distr_cls(laplace.Laplace)

  @parameterized.named_parameters(
      ('1d std laplace', (0, 1)),
      ('2d std laplace', (np.zeros(2), np.ones(2))),
      ('rank 2 std normal', (np.zeros((3, 2)), np.ones((3, 2)))),
      ('broadcasted loc', (0, np.ones(3))),
      ('broadcasted scale', (np.ones(3), 1)),
  )
  def test_event_shape(self, distr_params):
    super()._test_event_shape(distr_params, dict())

  @chex.all_variants
  @parameterized.named_parameters(
      ('1d std laplace, no shape', (0, 1), ()),
      ('1d std laplace, int shape', (0, 1), 1),
      ('1d std laplace, 1-tuple shape', (0, 1), (1,)),
      ('1d std laplace, 2-tuple shape', (0, 1), (2, 2)),
      ('2d std laplace, no shape', (np.zeros(2), np.ones(2)), ()),
      ('2d std laplace, int shape', ([0, 0], [1, 1]), 1),
      ('2d std laplace, 1-tuple shape', (np.zeros(2), np.ones(2)), (1,)),
      ('2d std laplace, 2-tuple shape', ([0, 0], [1, 1]), (2, 2)),
      ('rank 2 std normal, 2-tuple shape', (np.zeros((3, 2)), np.ones(
          (3, 2))), (2, 2)),
      ('broadcasted loc', (0, np.ones(3)), (2, 2)),
      ('broadcasted scale', (np.ones(3), 1), ()),
  )
  def test_sample_shape(self, distr_params, sample_shape):
    distr_params = (np.asarray(distr_params[0], dtype=np.float32),
                    np.asarray(distr_params[1], dtype=np.float32))
    super()._test_sample_shape(distr_params, dict(), sample_shape)

  @chex.all_variants
  @parameterized.named_parameters(
      ('float32', jnp.float32),
      ('float64', jnp.float64))
  def test_sample_dtype(self, dtype):
    with jax.experimental.enable_x64(dtype.dtype.itemsize == 8):
      dist = self.distrax_cls(
          loc=jnp.zeros((), dtype), scale=jnp.ones((), dtype))
      samples = self.variant(dist.sample)(seed=self.key)
      self.assertEqual(samples.dtype, dist.dtype)
      chex.assert_type(samples, dtype)

  @chex.all_variants
  @parameterized.named_parameters(
      ('1d std laplace, no shape', (0, 1), ()),
      ('1d std laplace, int shape', (0, 1), 1),
      ('1d std laplace, 1-tuple shape', (0, 1), (1,)),
      ('1d std laplace, 2-tuple shape', (0, 1), (2, 2)),
      ('2d std laplace, no shape', (np.zeros(2), np.ones(2)), ()),
      ('2d std laplace, int shape', ([0, 0], [1, 1]), 1),
      ('2d std laplace, 1-tuple shape', (np.zeros(2), np.ones(2)), (1,)),
      ('2d std laplace, 2-tuple shape', ([0, 0], [1, 1]), (2, 2)),
      ('rank 2 std normal, 2-tuple shape', (np.zeros((3, 2)), np.ones(
          (3, 2))), (2, 2)),
      ('broadcasted loc', (0, np.ones(3)), (2, 2)),
      ('broadcasted scale', (np.ones(3), 1), ()),
  )
  def test_sample_and_log_prob(self, distr_params, sample_shape):
    distr_params = (np.asarray(distr_params[0], dtype=np.float32),
                    np.asarray(distr_params[1], dtype=np.float32))
    super()._test_sample_and_log_prob(
        dist_args=distr_params,
        dist_kwargs=dict(),
        sample_shape=sample_shape,
        assertion_fn=self.assertion_fn(rtol=2e-2))

  @chex.all_variants
  @parameterized.named_parameters(
      ('1d dist, 1d value', (0, 1), 1),
      ('1d dist, 2d value', (0.5, 0.1), np.array([1, 2])),
      ('1d dist, 2d value as list', (0.5, 0.1), [1, 2]),
      ('2d dist, 1d value', (0.5 + np.zeros(2), 0.3 * np.ones(2)), 1),
      ('2d broadcasted dist, 1d value', (np.zeros(2), 0.8), 1),
      ('2d dist, 2d value', ([0.1, -0.5], 0.9 * np.ones(2)), np.array([1, 2])),
      ('1d dist, 1d value, edge case', (0, 1), 200),
  )
  def test_log_prob(self, distr_params, value):
    distr_params = (np.asarray(distr_params[0], dtype=np.float32),
                    np.asarray(distr_params[1], dtype=np.float32))
    value = np.asarray(value, dtype=np.float32)
    super()._test_attribute(
        attribute_string='log_prob',
        dist_args=distr_params,
        call_args=(value,),
        assertion_fn=self.assertion_fn(rtol=2e-2))

  @chex.all_variants
  @parameterized.named_parameters(
      ('1d dist, 1d value', (0, 1), 1),
      ('1d dist, 2d value', (0.5, 0.1), np.array([1, 2])),
      ('1d dist, 2d value as list', (0.5, 0.1), [1, 2]),
      ('2d dist, 1d value', (0.5 + np.zeros(2), 0.3 * np.ones(2)), 1),
      ('2d broadcasted dist, 1d value', (np.zeros(2), 0.8), 1),
      ('2d dist, 2d value', ([0.1, -0.5], 0.9 * np.ones(2)), np.array([1, 2])),
      ('1d dist, 1d value, edge case', (0, 1), 200),
  )
  def test_prob(self, distr_params, value):
    distr_params = (np.asarray(distr_params[0], dtype=np.float32),
                    np.asarray(distr_params[1], dtype=np.float32))
    value = np.asarray(value, dtype=np.float32)
    super()._test_attribute(
        attribute_string='prob',
        dist_args=distr_params,
        call_args=(value,),
        assertion_fn=self.assertion_fn(rtol=2e-2))

  @chex.all_variants
  @parameterized.named_parameters(
      ('1d dist, 1d value', (0, 1), 1),
      ('1d dist, 2d value', (0.5, 0.1), np.array([1, 2])),
      ('1d dist, 2d value as list', (0.5, 0.1), [1, 2]),
      ('2d dist, 1d value', (0.5 + np.zeros(2), 0.3 * np.ones(2)), 1),
      ('2d broadcasted dist, 1d value', (np.zeros(2), 0.8), 1),
      ('2d dist, 2d value', ([0.1, -0.5], 0.9 * np.ones(2)), np.array([1, 2])),
      ('1d dist, 1d value, edge case', (0, 1), 200),
  )
  def test_cdf(self, distr_params, value):
    distr_params = (np.asarray(distr_params[0], dtype=np.float32),
                    np.asarray(distr_params[1], dtype=np.float32))
    value = np.asarray(value, dtype=np.float32)
    super()._test_attribute(
        attribute_string='cdf',
        dist_args=distr_params,
        call_args=(value,),
        assertion_fn=self.assertion_fn(rtol=2e-2))

  @chex.all_variants
  @parameterized.named_parameters(
      ('1d dist, 1d value', (0, 1), 1),
      ('1d dist, 2d value', (0.5, 0.1), np.array([1, 2])),
      ('1d dist, 2d value as list', (0.5, 0.1), [1, 2]),
      ('2d dist, 1d value', (0.5 + np.zeros(2), 0.3 * np.ones(2)), 1),
      ('2d broadcasted dist, 1d value', (np.zeros(2), 0.8), 1),
      ('2d dist, 2d value', ([0.1, -0.5], 0.9 * np.ones(2)), np.array([1, 2])),
      ('1d dist, 1d value, edge case', (0, 1), 200),
  )
  def test_log_cdf(self, distr_params, value):
    distr_params = (np.asarray(distr_params[0], dtype=np.float32),
                    np.asarray(distr_params[1], dtype=np.float32))
    value = np.asarray(value, dtype=np.float32)
    super()._test_attribute(
        attribute_string='log_cdf',
        dist_args=distr_params,
        call_args=(value,),
        assertion_fn=self.assertion_fn(rtol=2e-2))

  @chex.all_variants
  @parameterized.named_parameters(
      ('1d dist, 1d value', (0, 1), 1),
      ('1d dist, 2d value', (0.5, 0.1), np.array([1, 2])),
      ('1d dist, 2d value as list', (0.5, 0.1), [1, 2]),
      ('2d dist, 1d value', (0.5 + np.zeros(2), 0.3 * np.ones(2)), 1),
      ('2d broadcasted dist, 1d value', (np.zeros(2), 0.8), 1),
      ('2d dist, 2d value', ([0.1, -0.5], 0.9 * np.ones(2)), np.array([1, 2])),
      ('1d dist, 1d value, edge case', (0, 1), -200),
  )
  def test_log_survival_function(self, distr_params, value):
    distr_params = (np.asarray(distr_params[0], dtype=np.float32),
                    np.asarray(distr_params[1], dtype=np.float32))
    value = np.asarray(value, dtype=np.float32)
    super()._test_attribute(
        attribute_string='log_survival_function',
        dist_args=distr_params,
        call_args=(value,),
        assertion_fn=self.assertion_fn(rtol=2e-2))

  @chex.all_variants(with_pmap=False)
  @parameterized.named_parameters(
      ('entropy', ([0., 1., -0.5], [0.5, 1., 1.5]), 'entropy'),
      ('entropy broadcasted loc', (0.5, [0.5, 1., 1.5]), 'entropy'),
      ('entropy broadcasted scale', ([0., 1., -0.5], 0.8), 'entropy'),
      ('mean', ([0., 1., -0.5], [0.5, 1., 1.5]), 'mean'),
      ('mean broadcasted loc', (0.5, [0.5, 1., 1.5]), 'mean'),
      ('mean broadcasted scale', ([0., 1., -0.5], 0.8), 'mean'),
      ('variance', ([0., 1., -0.5], [0.5, 1., 1.5]), 'variance'),
      ('variance broadcasted loc', (0.5, [0.5, 1., 1.5]), 'variance'),
      ('variance broadcasted scale', ([0., 1., -0.5], 0.8), 'variance'),
      ('stddev', ([0., 1., -0.5], [0.5, 1., 1.5]), 'stddev'),
      ('stddev broadcasted loc', (0.5, [0.5, 1., 1.5]), 'stddev'),
      ('stddev broadcasted scale', ([0., 1., -0.5], 0.8), 'stddev'),
      ('mode', ([0., 1., -0.5], [0.5, 1., 1.5]), 'mode'),
      ('mode broadcasted loc', (0.5, [0.5, 1., 1.5]), 'mode'),
      ('mode broadcasted scale', ([0., 1., -0.5], 0.8), 'mode'),
  )
  def test_method(self, distr_params, function_string):
    distr_params = (np.asarray(distr_params[0], dtype=np.float32),
                    np.asarray(distr_params[1], dtype=np.float32))
    super()._test_attribute(
        attribute_string=function_string,
        dist_args=distr_params,
        assertion_fn=self.assertion_fn(rtol=2e-2))

  @chex.all_variants(with_pmap=False)
  @parameterized.named_parameters(
      ('no broadcast', ([0., 1., -0.5], [0.5, 1., 1.5])),
      ('broadcasted loc', (0.5, [0.5, 1., 1.5])),
      ('broadcasted scale', ([0., 1., -0.5], 0.8)),
  )
  def test_median(self, distr_params):
    distr_params = (np.asarray(distr_params[0], dtype=np.float32),
                    np.asarray(distr_params[1], dtype=np.float32))
    dist = self.distrax_cls(*distr_params)
    self.assertion_fn(rtol=2e-2)(self.variant(dist.median)(), dist.mean())

  @chex.all_variants(with_pmap=False)
  @parameterized.named_parameters(
      ('kl distrax_to_distrax', 'kl_divergence', 'distrax_to_distrax'),
      ('kl distrax_to_tfp', 'kl_divergence', 'distrax_to_tfp'),
      ('kl tfp_to_distrax', 'kl_divergence', 'tfp_to_distrax'),
      ('cross-ent distrax_to_distrax', 'cross_entropy', 'distrax_to_distrax'),
      ('cross-ent distrax_to_tfp', 'cross_entropy', 'distrax_to_tfp'),
      ('cross-ent tfp_to_distrax', 'cross_entropy', 'tfp_to_distrax')
  )
  def test_with_two_distributions(self, function_string, mode_string):
    rng = np.random.default_rng(42)
    super()._test_with_two_distributions(
        attribute_string=function_string,
        mode_string=mode_string,
        dist1_kwargs={
            'loc': rng.normal(size=(4, 1, 2)),
            'scale': np.array([[0.8, 0.2], [0.1, 1.2], [1.4, 3.1]]),
        },
        dist2_kwargs={
            'loc': rng.normal(size=(3, 2)),
            'scale': 0.1 + rng.uniform(size=(4, 1, 2)),
        },
        assertion_fn=self.assertion_fn(rtol=2e-2))

  def test_jitable(self):
    super()._test_jittable((0., 1.))

  @parameterized.named_parameters(
      ('single element', 2),
      ('range', slice(-1)),
      ('range_2', (slice(None), slice(-1))),
      ('ellipsis', (Ellipsis, -1)),
  )
  def test_slice(self, slice_):
    rng = np.random.default_rng(42)
    loc = jnp.array(rng.normal(size=(3, 4, 5)))
    scale = jnp.array(rng.uniform(size=(3, 4, 5)))
    dist = self.distrax_cls(loc=loc, scale=scale)
    self.assertion_fn(rtol=2e-2)(dist[slice_].mean(), loc[slice_])

  def test_slice_different_parameterization(self):
    rng = np.random.default_rng(42)
    loc = jnp.array(rng.normal(size=(4,)))
    scale = jnp.array(rng.uniform(size=(3, 4)))
    dist = self.distrax_cls(loc=loc, scale=scale)
    self.assertion_fn(rtol=2e-2)(dist[0].loc, loc)  # Not slicing loc.
    self.assertion_fn(rtol=2e-2)(dist[0].scale, scale[0])


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for `kl_divergence` across different types of MultivariateNormal."""

from absl.testing import absltest
from absl.testing import parameterized

import chex

from distrax._src.distributions.mvn_diag import MultivariateNormalDiag
from distrax._src.distributions.mvn_diag_plus_low_rank import MultivariateNormalDiagPlusLowRank
from distrax._src.distributions.mvn_full_covariance import MultivariateNormalFullCovariance
from distrax._src.distributions.mvn_tri import MultivariateNormalTri

import numpy as np


def _get_dist_params(dist, batch_shape, dim, rng):
  """Generates random parameters depending on the distribution type."""
  if dist is MultivariateNormalDiag:
    distrax_dist_params = {
        'scale_diag': rng.normal(size=batch_shape + (dim,)),
    }
    tfp_dist_params = distrax_dist_params
  elif dist is MultivariateNormalDiagPlusLowRank:
    scale_diag = rng.normal(size=batch_shape + (dim,))
    scale_u_matrix = 0.2 * rng.normal(size=batch_shape + (dim, 2))
    scale_perturb_diag = rng.normal(size=batch_shape + (2,))
    scale_v_matrix = scale_u_matrix * np.expand_dims(
        scale_perturb_diag, axis=-2)
    distrax_dist_params = {
        'scale_diag': scale_diag,
        'scale_u_matrix': scale_u_matrix,
        'scale_v_matrix': scale_v_matrix,
    }
    tfp_dist_params = {
        'scale_diag': scale_diag,
        'scale_perturb_factor': scale_u_matrix,
        'scale_perturb_diag': scale_perturb_diag,
    }
  elif dist is MultivariateNormalTri:
    scale_tril = rng.normal(size=batch_shape + (dim, dim))
    distrax_dist_params = {
        'scale_tri': scale_tril,
        'is_lower': True,
    }
    tfp_dist_params = {
        'scale_tril': scale_tril,
    }
  elif dist is MultivariateNormalFullCovariance:
    matrix = rng.normal(size=batch_shape + (dim, dim))
    matrix_t = np.vectorize(np.transpose, signature='(k,k)->(k,k)')(matrix)
    distrax_dist_params = {
        'covariance_matrix': np.matmul(matrix, matrix_t),
    }
    tfp_dist_params = distrax_dist_params
  else:
    raise ValueError(f'Unsupported distribution type: {dist}')
  loc = rng.normal(size=batch_shape + (dim,))
  distrax_dist_params.update({'loc': loc})
  tfp_dist_params.update({'loc': loc})
  return distrax_dist_params, tfp_dist_params


class MultivariateNormalKLTest(parameterized.TestCase):

  @chex.all_variants(with_pmap=False)
  @parameterized.named_parameters(
      ('Diag vs DiagPlusLowRank',
       MultivariateNormalDiag, MultivariateNormalDiagPlusLowRank),
      ('Diag vs FullCovariance',
       MultivariateNormalDiag, MultivariateNormalFullCovariance),
      ('Diag vs Tri',
       MultivariateNormalDiag, MultivariateNormalTri),
      ('DiagPlusLowRank vs FullCovariance',
       MultivariateNormalDiagPlusLowRank, MultivariateNormalFullCovariance),
      ('DiagPlusLowRank vs Tri',
       MultivariateNormalDiagPlusLowRank, MultivariateNormalTri),
      ('Tri vs FullCovariance',
       MultivariateNormalTri, MultivariateNormalFullCovariance),
  )
  def test_two_distributions(self, dist1_type, dist2_type):
    rng = np.random.default_rng(42)

    distrax_dist1_params, tfp_dist1_params = _get_dist_params(
        dist1_type, batch_shape=(8, 1), dim=5, rng=rng)
    distrax_dist2_params, tfp_dist2_params = _get_dist_params(
        dist2_type, batch_shape=(6,), dim=5, rng=rng)

    dist1_distrax = dist1_type(**distrax_dist1_params)
    dist1_tfp = dist1_type.equiv_tfp_cls(**tfp_dist1_params)
    dist2_distrax = dist2_type(**distrax_dist2_params)
    dist2_tfp = dist2_type.equiv_tfp_cls(**tfp_dist2_params)

    for method in ['kl_divergence', 'cross_entropy']:
      expected_result1 = getattr(dist1_tfp, method)(dist2_tfp)
      expected_result2 = getattr(dist2_tfp, method)(dist1_tfp)
      for mode in ['distrax_to_distrax', 'distrax_to_tfp', 'tfp_to_distrax']:
        with self.subTest(method=method, mode=mode):
          if mode == 'distrax_to_distrax':
            result1 = self.variant(getattr(dist1_distrax, method))(
                dist2_distrax)
            result2 = self.variant(getattr(dist2_distrax, method))(
                dist1_distrax)
          elif mode == 'distrax_to_tfp':
            result1 = self.variant(getattr(dist1_distrax, method))(dist2_tfp)
            result2 = self.variant(getattr(dist2_distrax, method))(dist1_tfp)
          elif mode == 'tfp_to_distrax':
            result1 = self.variant(getattr(dist1_tfp, method))(dist2_distrax)
            result2 = self.variant(getattr(dist2_tfp, method))(dist1_distrax)
          else:
            raise ValueError(f'Unsupported mode: {mode}')
          np.testing.assert_allclose(result1, expected_result1, rtol=0.03)
          np.testing.assert_allclose(result2, expected_result2, rtol=0.03)


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for `bijector.py`."""

from absl.testing import absltest
from absl.testing import parameterized

import chex
from distrax._src.bijectors import bijector
import jax
import jax.numpy as jnp
import numpy as np


class DummyBijector(bijector.Bijector):

  def forward_and_log_det(self, x):
    super()._check_forward_input_shape(x)
    return x, jnp.zeros(x.shape[:-1], jnp.float_)

  def inverse_and_log_det(self, y):
    super()._check_inverse_input_shape(y)
    return y, jnp.zeros(y.shape[:-1], jnp.float_)


class BijectorTest(parameterized.TestCase):

  @parameterized.named_parameters(
      ('negative ndims_in', -1, 1, False, False),
      ('negative ndims_out', 1, -1, False, False),
      ('non-consistent constant properties', 1, 1, True, False),
  )
  def test_invalid_parameters(self, ndims_in, ndims_out, cnst_jac, cnst_logdet):
    with self.assertRaises(ValueError):
      DummyBijector(ndims_in, ndims_out, cnst_jac, cnst_logdet)

  @chex.all_variants
  @parameterized.parameters('forward', 'inverse')
  def test_invalid_inputs(self, method_str):
    bij = DummyBijector(1, 1, True, True)
    fn = self.variant(getattr(bij, method_str))
    with self.assertRaises(ValueError):
      fn(jnp.zeros(()))

  def test_jittable(self):
    @jax.jit
    def forward(bij, x):
      return bij.forward(x)

    bij = DummyBijector(1, 1, True, True)
    x = jnp.zeros((4,))
    np.testing.assert_allclose(forward(bij, x), x)


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Distribution abstract base class."""

import abc
import collections.abc
import contextlib
import functools
import operator
import typing

from typing import Any, Generic, Iterable, Mapping, Sequence, Tuple, TypeVar, Union

import chex
from distrax._src.utils import jittable
import jax
import jax.numpy as jnp
import numpy as np
from tensorflow_probability.substrates import jax as tfp


tfd = tfp.distributions

Array = chex.Array
PRNGKey = chex.PRNGKey
IntLike = Union[int, np.int16, np.int32, np.int64]
# Generic type.
T = TypeVar('T')
# Generic nested type.
NestedT = Union[T, Iterable['NestedT'], Mapping[Any, 'NestedT']]  # pylint: disable=invalid-name
# Nested types.
EventT = TypeVar('EventT', bound=NestedT[Array])
ShapeT = TypeVar('ShapeT', bound=NestedT[Tuple[int, ...]])
DTypeT = TypeVar('DTypeT', bound=NestedT[jnp.dtype])


class Distribution(
    jittable.Jittable, Generic[EventT, ShapeT, DTypeT], metaclass=abc.ABCMeta):
  """Jittable abstract base class for all Distrax distributions."""

  @abc.abstractmethod
  def _sample_n(self, key: PRNGKey, n: int) -> EventT:
    """Returns `n` samples."""

  def _sample_n_and_log_prob(
      self,
      key: PRNGKey,
      n: int,
  ) -> Tuple[EventT, Array]:
    """Returns `n` samples and their log probs.

    By default, it just calls `log_prob` on the generated samples. However, for
    many distributions it's more efficient to compute the log prob of samples
    than of arbitrary events (for example, there's no need to check that a
    sample is within the distribution's domain). If that's the case, a subclass
    may override this method with a more efficient implementation.

    Args:
      key: PRNG key.
      n: Number of samples to generate.

    Returns:
      A tuple of `n` samples and their log probs.
    """
    samples = self._sample_n(key, n)
    log_prob = self.log_prob(samples)
    return samples, log_prob

  @abc.abstractmethod
  def log_prob(self, value: EventT) -> Array:
    """Calculates the log probability of an event.

    Args:
      value: An event.

    Returns:
      The log probability log P(value).
    """

  def prob(self, value: EventT) -> Array:
    """Calculates the probability of an event.

    Args:
      value: An event.

    Returns:
      The probability P(value).
    """
    return jnp.exp(self.log_prob(value))

  @property
  @abc.abstractmethod
  def event_shape(self) -> ShapeT:
    """Shape of event of distribution samples."""

  @property
  def batch_shape(self) -> Tuple[int, ...]:
    """Shape of batch of distribution samples."""
    sample_spec = jax.eval_shape(
        self.sample, seed=jax.random.PRNGKey(0), sample_shape=())

    if not self.event_shape:
      batch_shapes = jax.tree_util.tree_map(lambda x: x.shape, sample_spec)
    else:
      batch_shapes = jax.tree_util.tree_map(
          lambda s, e: s.shape[:s.ndim - len(e)], sample_spec, self.event_shape)

    # Get flat batch shapes.
    batch_shapes = jax.tree_util.tree_structure(sample_spec).flatten_up_to(
        batch_shapes)
    if not batch_shapes:
      return ()

    # Ensure batch shapes are consistent.
    batch_shape = batch_shapes[0]
    for i in range(1, len(batch_shapes)):
      np.testing.assert_equal(batch_shape, batch_shapes[i])
    return batch_shape

  @property
  def name(self) -> str:
    """Distribution name."""
    return type(self).__name__

  @property
  def dtype(self) -> DTypeT:
    """The data type of the samples generated by the distribution."""
    sample_spec = jax.eval_shape(
        self.sample, seed=jax.random.PRNGKey(0), sample_shape=())
    return jax.tree_util.tree_map(lambda x: x.dtype, sample_spec)

  def sample(self,
             *,
             seed: Union[IntLike, PRNGKey],
             sample_shape: Union[IntLike, Sequence[IntLike]] = ()) -> EventT:
    """Samples an event.

    Args:
      seed: PRNG key or integer seed.
      sample_shape: Additional leading dimensions for sample.

    Returns:
      A sample of shape `sample_shape + self.batch_shape + self.event_shape`.
    """
    rng, sample_shape = convert_seed_and_sample_shape(seed, sample_shape)
    num_samples = functools.reduce(operator.mul, sample_shape, 1)  # product

    samples = self._sample_n(rng, num_samples)
    return jax.tree_util.tree_map(
        lambda t: t.reshape(sample_shape + t.shape[1:]), samples)

  def sample_and_log_prob(
      self,
      *,
      seed: Union[IntLike, PRNGKey],
      sample_shape: Union[IntLike, Sequence[IntLike]] = ()
  ) -> Tuple[EventT, Array]:
    """Returns a sample and associated log probability. See `sample`."""
    rng, sample_shape = convert_seed_and_sample_shape(seed, sample_shape)
    num_samples = functools.reduce(operator.mul, sample_shape, 1)  # product

    samples, log_prob = self._sample_n_and_log_prob(rng, num_samples)
    samples, log_prob = jax.tree_util.tree_map(
        lambda t: t.reshape(sample_shape + t.shape[1:]), (samples, log_prob))
    return samples, log_prob

  def kl_divergence(self, other_dist, **kwargs) -> Array:
    """Calculates the KL divergence to another distribution.

    Args:
      other_dist: A compatible Distax or TFP Distribution.
      **kwargs: Additional kwargs.

    Returns:
      The KL divergence `KL(self || other_dist)`.
    """
    return tfd.kullback_leibler.kl_divergence(self, other_dist, **kwargs)

  def entropy(self) -> Array:
    """Calculates the Shannon entropy (in nats)."""
    raise NotImplementedError(
        f'Distribution `{self.name}` does not implement `entropy`.')

  def log_cdf(self, value: EventT) -> Array:
    """Evaluates the log cumulative distribution function at `value`.

    Args:
      value: An event.

    Returns:
      The log CDF evaluated at value, i.e. log P[X <= value].
    """
    raise NotImplementedError(
        f'Distribution `{self.name}` does not implement `log_cdf`.')

  def cdf(self, value: EventT) -> Array:
    """Evaluates the cumulative distribution function at `value`.

    Args:
      value: An event.

    Returns:
      The CDF evaluated at value, i.e. P[X <= value].
    """
    return jnp.exp(self.log_cdf(value))

  def survival_function(self, value: EventT) -> Array:
    """Evaluates the survival function at `value`.

    Note that by default we use a numerically not necessarily stable definition
    of the survival function in terms of the CDF.
    More stable definitions should be implemented in subclasses for
    distributions for which they exist.

    Args:
      value: An event.

    Returns:
      The survival function evaluated at `value`, i.e. P[X > value]
    """
    if not self.event_shape:
      # Defined for univariate distributions only.
      return 1. - self.cdf(value)
    else:
      raise NotImplementedError('`survival_function` is not defined for '
                                f'distribution `{self.name}`.')

  def log_survival_function(self, value: EventT) -> Array:
    """Evaluates the log of the survival function at `value`.

    Note that by default we use a numerically not necessarily stable definition
    of the log of the survival function in terms of the CDF.
    More stable definitions should be implemented in subclasses for
    distributions for which they exist.

    Args:
      value: An event.

    Returns:
      The log of the survival function evaluated at `value`, i.e.
      log P[X > value]
    """
    if not self.event_shape:
      # Defined for univariate distributions only.
      return jnp.log1p(-self.cdf(value))
    else:
      raise NotImplementedError('`log_survival_function` is not defined for '
                                f'distribution `{self.name}`.')

  def mean(self) -> EventT:
    """Calculates the mean."""
    raise NotImplementedError(
        f'Distribution `{self.name}` does not implement `mean`.')

  def median(self) -> EventT:
    """Calculates the median."""
    raise NotImplementedError(
        f'Distribution `{self.name}` does not implement `median`.')

  def variance(self) -> EventT:
    """Calculates the variance."""
    raise NotImplementedError(
        f'Distribution `{self.name}` does not implement `variance`.')

  def stddev(self) -> EventT:
    """Calculates the standard deviation."""
    return jnp.sqrt(self.variance())

  def mode(self) -> EventT:
    """Calculates the mode."""
    raise NotImplementedError(
        f'Distribution `{self.name}` does not implement `mode`.')

  def cross_entropy(self, other_dist, **kwargs) -> Array:
    """Calculates the cross entropy to another distribution.

    Args:
      other_dist: A compatible Distax or TFP Distribution.
      **kwargs: Additional kwargs.

    Returns:
      The cross entropy `H(self || other_dist)`.
    """
    return self.kl_divergence(other_dist, **kwargs) + self.entropy()

  @contextlib.contextmanager
  def _name_and_control_scope(self, *unused_a, **unused_k):
    yield

  def __getitem__(self, index) -> 'Distribution':
    """Returns a matching distribution obtained by indexing the batch shape.

    Args:
      index: An object, typically int or slice (or a tuple thereof), used for
        indexing the distribution.
    """
    raise NotImplementedError(f'Indexing not implemented for `{self.name}`.')


def convert_seed_and_sample_shape(
    seed: Union[IntLike, PRNGKey],
    sample_shape: Union[IntLike, Sequence[IntLike]]
) -> Tuple[PRNGKey, Tuple[int, ...]]:
  """Shared functionality to ensure that seeds and shapes are the right type."""

  if not isinstance(sample_shape, collections.abc.Sequence):
    sample_shape = (sample_shape,)
  sample_shape = tuple(map(int, sample_shape))

  if isinstance(seed, (int, np.signedinteger)):
    rng = jax.random.PRNGKey(seed)
  else:  # key is of type PRNGKey
    rng = seed

  return rng, sample_shape  # type: ignore[bad-return-type]


def to_batch_shape_index(
    batch_shape: Tuple[int, ...],
    index,
) -> Tuple[jnp.ndarray, ...]:
  """Utility function that transforms the index to respect the batch shape.

  When indexing a distribution we only want to index based on the batch shape.
  For example, a Categorical with logits shaped (2, 3, 4) has batch shape of
  (2, 3) and number of categoricals 4. Indexing this distribution creates a new
  distribution with indexed logits. If the index is [0], the new distribution's
  logits will be shaped (3, 4). But if the index is [..., -1] the new logits
  should be shaped (2, 4), but applying the index operation on logits directly
  will result in shape (2, 3). This function fixes such indices such that they
  are only applied on the batch shape.

  Args:
    batch_shape: Distribution's batch_shape.
    index: An object, typically int or slice (or a tuple thereof), used for
      indexing the distribution.

  Returns:
    A new index that is only applied on the batch shape.
  """
  try:
    new_index = [x[index] for x in np.indices(batch_shape)]
    return tuple(new_index)
  except IndexError as e:
    raise IndexError(f'Batch shape `{batch_shape}` not compatible with index '
                     f'`{index}`.') from e


DistributionLike = Union[Distribution, tfd.Distribution]
DistributionT = typing.TypeVar('DistributionT', bound=Distribution)



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for `bijector_from_tfp.py`."""

from absl.testing import absltest
from absl.testing import parameterized

import chex
from distrax._src.bijectors import bijector_from_tfp
import jax
import jax.numpy as jnp
import numpy as np
from tensorflow_probability.substrates import jax as tfp

tfb = tfp.bijectors


class BijectorFromTFPTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    bjs = {}
    bjs['BatchedChain'] = tfb.Chain([
        tfb.Shift(jnp.zeros((4, 2, 3))),
        tfb.ScaleMatvecDiag([[1., 2., 3.], [4., 5., 6.]])
    ])
    bjs['Square'] = tfb.Square()
    bjs['ScaleScalar'] = tfb.Scale(2.)
    bjs['ScaleMatrix'] = tfb.Scale(2. * jnp.ones((3, 2)))
    bjs['Reshape'] = tfb.Reshape((2, 3), (6,))

    # To parallelize pytest runs.
    # See https://github.com/pytest-dev/pytest-xdist/issues/432.
    for name, bij in bjs.items():
      bij.__repr__ = lambda _, name_=name: name_

    self._test_bijectors = bjs

  @chex.all_variants
  @parameterized.parameters(
      ('Square', (), (), (), ()),
      ('Square', (2, 3), (), (2, 3), ()),
      ('ScaleScalar', (), (), (), ()),
      ('ScaleScalar', (2, 3), (), (2, 3), ()),
      ('ScaleMatrix', (), (), (3, 2), ()),
      ('ScaleMatrix', (2,), (), (3, 2), ()),
      ('ScaleMatrix', (1, 1), (), (3, 2), ()),
      ('ScaleMatrix', (4, 1, 1), (), (4, 3, 2), ()),
      ('ScaleMatrix', (4, 3, 2), (), (4, 3, 2), ()),
      ('Reshape', (), (6,), (), (2, 3)),
      ('Reshape', (10,), (6,), (10,), (2, 3)),
      ('BatchedChain', (), (3,), (4, 2), (3,)),
      ('BatchedChain', (2,), (3,), (4, 2), (3,)),
      ('BatchedChain', (4, 1), (3,), (4, 2), (3,)),
      ('BatchedChain', (5, 1, 2), (3,), (5, 4, 2), (3,)),
  )
  def test_forward_methods_are_correct(self, tfp_bij_name, batch_shape_in,
                                       event_shape_in, batch_shape_out,
                                       event_shape_out):
    tfp_bij = self._test_bijectors[tfp_bij_name]
    bij = bijector_from_tfp.BijectorFromTFP(tfp_bij)
    key = jax.random.PRNGKey(42)
    x = jax.random.uniform(key, batch_shape_in + event_shape_in)

    y = self.variant(bij.forward)(x)
    logdet = self.variant(bij.forward_log_det_jacobian)(x)
    y_tfp = tfp_bij.forward(x)
    logdet_tfp = tfp_bij.forward_log_det_jacobian(x, len(event_shape_in))
    logdet_tfp = jnp.broadcast_to(logdet_tfp, batch_shape_out)

    self.assertEqual(y.shape, batch_shape_out + event_shape_out)
    self.assertEqual(logdet.shape, batch_shape_out)
    np.testing.assert_allclose(y, y_tfp, atol=1e-8)
    np.testing.assert_allclose(logdet, logdet_tfp, atol=1e-4)

  @chex.all_variants
  @parameterized.parameters(
      ('Square', (), (), (), ()),
      ('Square', (2, 3), (), (2, 3), ()),
      ('ScaleScalar', (), (), (), ()),
      ('ScaleScalar', (2, 3), (), (2, 3), ()),
      ('ScaleMatrix', (3, 2), (), (), ()),
      ('ScaleMatrix', (3, 2), (), (2,), ()),
      ('ScaleMatrix', (3, 2), (), (1, 1), ()),
      ('ScaleMatrix', (4, 3, 2), (), (4, 1, 1), ()),
      ('ScaleMatrix', (4, 3, 2), (), (4, 3, 2), ()),
      ('Reshape', (), (6,), (), (2, 3)),
      ('Reshape', (10,), (6,), (10,), (2, 3)),
      ('BatchedChain', (4, 2), (3,), (), (3,)),
      ('BatchedChain', (4, 2), (3,), (2,), (3,)),
      ('BatchedChain', (4, 2), (3,), (4, 1), (3,)),
      ('BatchedChain', (5, 4, 2), (3,), (5, 1, 2), (3,)),
  )
  def test_inverse_methods_are_correct(self, tfp_bij_name, batch_shape_in,
                                       event_shape_in, batch_shape_out,
                                       event_shape_out):
    tfp_bij = self._test_bijectors[tfp_bij_name]
    bij = bijector_from_tfp.BijectorFromTFP(tfp_bij)
    key = jax.random.PRNGKey(42)
    y = jax.random.uniform(key, batch_shape_out + event_shape_out)

    x = self.variant(bij.inverse)(y)
    logdet = self.variant(bij.inverse_log_det_jacobian)(y)
    x_tfp = tfp_bij.inverse(y)
    logdet_tfp = tfp_bij.inverse_log_det_jacobian(y, len(event_shape_out))
    logdet_tfp = jnp.broadcast_to(logdet_tfp, batch_shape_in)

    self.assertEqual(x.shape, batch_shape_in + event_shape_in)
    self.assertEqual(logdet.shape, batch_shape_in)
    np.testing.assert_allclose(x, x_tfp, atol=1e-8)
    np.testing.assert_allclose(logdet, logdet_tfp, atol=1e-4)

  @chex.all_variants
  @parameterized.parameters(
      ('Square', (), (), (), ()),
      ('Square', (2, 3), (), (2, 3), ()),
      ('ScaleScalar', (), (), (), ()),
      ('ScaleScalar', (2, 3), (), (2, 3), ()),
      ('ScaleMatrix', (), (), (), ()),
      ('ScaleMatrix', (2,), (), (2,), ()),
      ('ScaleMatrix', (1, 1), (), (1, 1), ()),
      ('ScaleMatrix', (4, 1, 1), (), (4, 1, 1), ()),
      ('ScaleMatrix', (4, 3, 2), (), (4, 3, 2), ()),
      ('Reshape', (), (6,), (), (2, 3)),
      ('Reshape', (10,), (6,), (10,), (2, 3)),
      ('BatchedChain', (), (3,), (), (3,)),
      ('BatchedChain', (2,), (3,), (2,), (3,)),
      ('BatchedChain', (4, 1), (3,), (4, 1), (3,)),
      ('BatchedChain', (5, 1, 2), (3,), (5, 1, 2), (3,)),
  )
  def test_composite_methods_are_consistent(self, tfp_bij_name, batch_shape_in,
                                            event_shape_in, batch_shape_out,
                                            event_shape_out):
    key1, key2 = jax.random.split(jax.random.PRNGKey(42))
    tfp_bij = self._test_bijectors[tfp_bij_name]
    bij = bijector_from_tfp.BijectorFromTFP(tfp_bij)

    # Forward methods.
    x = jax.random.uniform(key1, batch_shape_in + event_shape_in)
    y1 = self.variant(bij.forward)(x)
    logdet1 = self.variant(bij.forward_log_det_jacobian)(x)
    y2, logdet2 = self.variant(bij.forward_and_log_det)(x)
    self.assertEqual(y1.shape, y2.shape)
    self.assertEqual(logdet1.shape, logdet2.shape)
    np.testing.assert_allclose(y1, y2, atol=1e-8)
    np.testing.assert_allclose(logdet1, logdet2, atol=1e-8)

    # Inverse methods.
    y = jax.random.uniform(key2, batch_shape_out + event_shape_out)
    x1 = self.variant(bij.inverse)(y)
    logdet1 = self.variant(bij.inverse_log_det_jacobian)(y)
    x2, logdet2 = self.variant(bij.inverse_and_log_det)(y)
    self.assertEqual(x1.shape, x2.shape)
    self.assertEqual(logdet1.shape, logdet2.shape)
    np.testing.assert_allclose(x1, x2, atol=1e-8)
    np.testing.assert_allclose(logdet1, logdet2, atol=1e-8)

  @chex.all_variants
  @parameterized.parameters(
      ('Square', (), (), (), ()),
      ('Square', (2, 3), (), (2, 3), ()),
      ('ScaleScalar', (), (), (), ()),
      ('ScaleScalar', (2, 3), (), (2, 3), ()),
      ('ScaleMatrix', (), (), (), ()),
      ('ScaleMatrix', (2,), (), (2,), ()),
      ('ScaleMatrix', (1, 1), (), (1, 1), ()),
      ('ScaleMatrix', (4, 1, 1), (), (4, 1, 1), ()),
      ('ScaleMatrix', (4, 3, 2), (), (4, 3, 2), ()),
      ('Reshape', (), (6,), (), (2, 3)),
      ('Reshape', (10,), (6,), (10,), (2, 3)),
      ('BatchedChain', (), (3,), (), (3,)),
      ('BatchedChain', (2,), (3,), (2,), (3,)),
      ('BatchedChain', (4, 1), (3,), (4, 1), (3,)),
      ('BatchedChain', (5, 1, 2), (3,), (5, 1, 2), (3,)),
  )
  def test_works_with_tfp_caching(self, tfp_bij_name, batch_shape_in,
                                  event_shape_in, batch_shape_out,
                                  event_shape_out):
    tfp_bij = self._test_bijectors[tfp_bij_name]
    bij = bijector_from_tfp.BijectorFromTFP(tfp_bij)
    key1, key2 = jax.random.split(jax.random.PRNGKey(42))

    # Forward caching.
    x = jax.random.uniform(key1, batch_shape_in + event_shape_in)
    y = self.variant(bij.forward)(x)
    x1 = self.variant(bij.inverse)(y)
    logdet1 = self.variant(bij.inverse_log_det_jacobian)(y)
    x2, logdet2 = self.variant(bij.inverse_and_log_det)(y)
    self.assertEqual(x1.shape, x2.shape)
    self.assertEqual(logdet1.shape, logdet2.shape)
    np.testing.assert_allclose(x1, x2, atol=1e-8)
    np.testing.assert_allclose(logdet1, logdet2, atol=1e-8)

    # Inverse caching.
    y = jax.random.uniform(key2, batch_shape_out + event_shape_out)
    x = self.variant(bij.inverse)(y)
    y1 = self.variant(bij.forward)(x)
    logdet1 = self.variant(bij.forward_log_det_jacobian)(x)
    y2, logdet2 = self.variant(bij.forward_and_log_det)(x)
    self.assertEqual(y1.shape, y2.shape)
    self.assertEqual(logdet1.shape, logdet2.shape)
    np.testing.assert_allclose(y1, y2, atol=1e-8)
    np.testing.assert_allclose(logdet1, logdet2, atol=1e-8)

  def test_access_properties_tfp_bijector(self):
    tfp_bij = self._test_bijectors['BatchedChain']
    bij = bijector_from_tfp.BijectorFromTFP(tfp_bij)
    # Access the attribute `bijectors`
    np.testing.assert_allclose(
        bij.bijectors[0].shift, tfp_bij.bijectors[0].shift, atol=1e-8)
    np.testing.assert_allclose(
        bij.bijectors[1].scale.diag, tfp_bij.bijectors[1].scale.diag, atol=1e-8)

  def test_jittable(self):

    @jax.jit
    def f(x, b):
      return b.forward(x)

    bijector = bijector_from_tfp.BijectorFromTFP(tfb.Tanh())
    x = np.zeros(())
    f(x, bijector)


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for `straight_through.py`."""

from absl.testing import absltest
from absl.testing import parameterized

import chex
from distrax._src.distributions import one_hot_categorical
from distrax._src.distributions import straight_through
from distrax._src.utils import equivalence
from distrax._src.utils import math
import jax
import jax.numpy as jnp
import numpy as np


class StraightThroughTest(equivalence.EquivalenceTest):

  def setUp(self):
    super().setUp()
    self._init_distr_cls(
        straight_through.straight_through_wrapper(
            one_hot_categorical.OneHotCategorical))

  @chex.all_variants
  @parameterized.named_parameters(
      ('1d logits, no shape', {'logits': [0.0, 1.0, -0.5]}, ()),
      ('1d probs, no shape', {'probs': [0.2, 0.5, 0.3]}, ()),
      ('1d logits, int shape', {'logits': [0.0, 1.0, -0.5]}, 1),
      ('1d probs, int shape', {'probs': [0.2, 0.5, 0.3]}, 1),
      ('1d logits, 1-tuple shape', {'logits': [0.0, 1.0, -0.5]}, (1,)),
      ('1d probs, 1-tuple shape', {'probs': [0.2, 0.5, 0.3]}, (1,)),
      ('1d logits, 2-tuple shape', {'logits': [0.0, 50., -0.5]}, (5, 4)),
      ('1d probs, 2-tuple shape', {'probs': [0.01, 0.99, 0.]}, (5, 4)),
      ('2d logits, no shape', {'logits': [[0.0, 1.0, -0.5],
                                          [-0.1, 0.3, 0.0]]}, ()),
      ('2d probs, no shape', {'probs': [[0.1, 0.4, 0.5],
                                        [0.5, 0.25, 0.25]]}, ()),
      ('2d logits, int shape', {'logits': [[0.0, 50.0, -0.5],
                                           [-0.1, -0.3, 0.2]]}, 4),
      ('2d probs, int shape', {'probs': [[0.005, 0.005, 0.99],
                                         [0.99, 0., 0.01]]}, 4),
      ('2d logits, 1-tuple shape', {'logits': [[0.0, 1.0, -0.5],
                                               [-0.1, 0.3, 200.0]]}, (5,)),
      ('2d probs, 1-tuple shape', {'probs': [[0., 0.01, 0.99],
                                             [0., 0.99, 0.01]]}, (5,)),
      ('2d logits, 2-tuple shape', {'logits': [[0.0, 1.0, -0.5],
                                               [-0.1, 0.3, 1000.0]]}, (5, 4)),
      ('2d probs, 2-tuple shape', {'probs': [[0.01, 0.99, 0.],
                                             [0.99, 0., 0.01]]}, (5, 4)),
  )
  def test_sample(self, dist_params, sample_shape):

    def loss(dist_params, dist_cls, sample_shape):
      """Loss on sample, used both for distrax and TFP."""

      # Sample.
      dist = dist_cls(**dist_params)
      sample_fn = dist.sample

      def sample_fn_wrapper(seed, sample_shape):
        """To test with pmap that requires positional arguments."""
        return sample_fn(seed=seed, sample_shape=sample_shape)

      if hasattr(self, 'variant'):
        sample_fn_wrapper = self.variant(static_argnums=(1,))(sample_fn_wrapper)
      sample = sample_fn_wrapper(self.key, sample_shape)
      return jnp.sum((sample)**2).astype(jnp.float32), sample

    # TFP softmax gradient.
    def straight_through_tfp_loss(dist_params, dist_cls, sample_shape):
      """Loss on a straight-through gradient of the tfp sample."""
      # Distrax normalises the distribution parameters. We want to make sure
      # that they are normalised for tfp too, or the gradient might differ.
      try:
        dist_params['logits'] = math.normalize(logits=dist_params['logits'])
      except KeyError:
        dist_params['probs'] = math.normalize(probs=dist_params['probs'])

      # Sample.
      dist = dist_cls(**dist_params)
      sample_fn = dist.sample

      def sample_fn_wrapper(seed, sample_shape):
        """To test with pmap that requires positional arguments."""
        return sample_fn(seed=seed, sample_shape=sample_shape)

      if hasattr(self, 'variant'):
        sample_fn_wrapper = self.variant(static_argnums=(1,))(sample_fn_wrapper)
      sample = sample_fn_wrapper(self.key, sample_shape)

      # Straight-through gradient.
      def _pad(probs, shape):
        if isinstance(shape, int):
          return probs
        while len(probs.shape) < len(shape):
          probs = probs[None]
        return probs
      probs = dist.probs_parameter()
      padded_probs = _pad(probs, sample_shape)
      sample += padded_probs - jax.lax.stop_gradient(padded_probs)

      return jnp.sum((sample)**2).astype(jnp.float32), sample

    # Straight-through gradient and sample.
    sample_grad, sample = jax.grad(loss, has_aux=True)(dist_params,
                                                       self.distrax_cls,
                                                       sample_shape)
    # TFP gradient (zero) and sample.
    tfp_sample_grad, tfp_sample = jax.grad(loss, has_aux=True)(dist_params,
                                                               self.tfp_cls,
                                                               sample_shape)
    # TFP straight-through gradient and sample.
    tfp_st_sample_grad, tfp_st_sample = jax.grad(straight_through_tfp_loss,
                                                 has_aux=True)(dist_params,
                                                               self.tfp_cls,
                                                               sample_shape)

    # TEST: the samples have the same size, and the straight-through gradient
    # doesn't affect the tfp sample.
    chex.assert_equal_shape((sample, tfp_sample))
    self.assertion_fn(rtol=2e-3)(tfp_sample, tfp_st_sample)
    # TEST: the TFP gradient is zero.
    assert (jnp.asarray(*tfp_sample_grad.values()) == 0).all()
    # TEST: the TFP straight-through gradient is non zero.
    assert (jnp.asarray(*tfp_st_sample_grad.values()) != 0).any()
    # Test that the TFP straight-through gradient is equal to the one from
    # distrax when the samples from distrax and tfp are the same (due to
    # stochasticity the samples can differ - we are using skewed distributions
    # on purpose in the parametrization of the test to make sure that the
    # samples match most of the time).
    sample_grad_v = jnp.stack(jnp.array(*sample_grad.values()))
    tfp_st_sample_grad_v = jnp.stack(jnp.array(*tfp_st_sample_grad.values()))
    if np.all(sample == tfp_st_sample):
      self.assertion_fn(rtol=2e-3)(sample_grad_v, tfp_st_sample_grad_v)


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""MultivariateNormalFromBijector distribution."""

from typing import Callable, Union

import chex

from distrax._src.bijectors import block
from distrax._src.bijectors import chain
from distrax._src.bijectors import diag_linear
from distrax._src.bijectors import linear
from distrax._src.bijectors import shift
from distrax._src.distributions import independent
from distrax._src.distributions import normal
from distrax._src.distributions import transformed

import jax
import jax.numpy as jnp

from tensorflow_probability.substrates import jax as tfp

tfd = tfp.distributions

Array = chex.Array


def _check_input_parameters_are_valid(scale: linear.Linear, loc: Array) -> None:
  """Raises an error if `scale` and `loc` are not valid."""
  if loc.ndim < 1:
    raise ValueError('`loc` must have at least 1 dimension.')
  if scale.event_dims != loc.shape[-1]:
    raise ValueError(
        f'`scale` and `loc` have inconsistent dimensionality: '
        f'`scale.event_dims = {scale.event_dims} and '
        f'`loc.shape[-1] = {loc.shape[-1]}.')


class MultivariateNormalFromBijector(transformed.Transformed):
  """Multivariate normal distribution on `R^k`.

  The multivariate normal over `x` is characterized by an invertible affine
  transformation `x = f(z) = A @ z + b`, where `z` is a random variable that
  follows a standard multivariate normal on `R^k`, i.e., `p(z) = N(0, I_k)`,
  `A` is a `k x k` transformation matrix, and `b` is a `k`-dimensional vector.

  The resulting PDF on `x` is a multivariate normal, `p(x) = N(b, C)`, where
  `C = A @ A.T` is the covariance matrix. Additional leading dimensions (if any)
  index batches.

  The transformation `x = f(z)` must be specified by a linear scale bijector
  implementing the operation `A @ z` and a shift (or location) term `b`.
  """

  def __init__(self, loc: Array, scale: linear.Linear):
    """Initializes the distribution.

    Args:
      loc: The term `b`, i.e., the mean of the multivariate normal distribution.
      scale: The bijector specifying the linear transformation `A @ z`, as
        described in the class docstring.
    """
    _check_input_parameters_are_valid(scale, loc)
    batch_shape = jnp.broadcast_shapes(scale.batch_shape, loc.shape[:-1])
    dtype = jnp.result_type(scale.dtype, loc.dtype)

    # Build a standard multivariate Gaussian with the right `batch_shape`.
    std_mvn_dist = independent.Independent(
        distribution=normal.Normal(
            loc=jnp.zeros(batch_shape + loc.shape[-1:], dtype=dtype),
            scale=1.),
        reinterpreted_batch_ndims=1)
    # Form the bijector `f(x) = Ax + b`.
    bijector = chain.Chain([block.Block(shift.Shift(loc), ndims=1), scale])
    super().__init__(distribution=std_mvn_dist, bijector=bijector)
    self._scale = scale
    self._loc = loc
    self._event_shape = loc.shape[-1:]
    self._batch_shape = batch_shape
    self._dtype = dtype

  @property
  def scale(self) -> linear.Linear:
    """The scale bijector."""
    return self._scale

  @property
  def loc(self) -> Array:
    """The `loc` parameter of the distribution."""
    shape = self.batch_shape + self.event_shape
    return jnp.broadcast_to(self._loc, shape=shape)

  def mean(self) -> Array:
    """Calculates the mean."""
    return self.loc

  def median(self) -> Array:
    """Calculates the median."""
    return self.loc

  def mode(self) -> Array:
    """Calculates the mode."""
    return self.loc

  def covariance(self) -> Array:
    """Calculates the covariance matrix.

    Unlike TFP, which would drop leading dimensions, in Distrax the covariance
    matrix always has shape `batch_shape + (num_dims, num_dims)`. This helps to
    keep things simple and predictable.

    Returns:
      The covariance matrix, of shape `k x k` (broadcasted to match the batch
      shape of the distribution).
    """
    if isinstance(self.scale, diag_linear.DiagLinear):
      result = jnp.vectorize(jnp.diag, signature='(k)->(k,k)')(self.variance())
    else:
      result = jax.vmap(self.scale.forward, in_axes=-2, out_axes=-2)(
          self._scale.matrix)
    return jnp.broadcast_to(
        result, self.batch_shape + self.event_shape + self.event_shape)

  def variance(self) -> Array:
    """Calculates the variance of all one-dimensional marginals."""
    if isinstance(self.scale, diag_linear.DiagLinear):
      result = jnp.square(self.scale.diag)
    else:
      scale_matrix = self._scale.matrix
      result = jnp.sum(scale_matrix * scale_matrix, axis=-1)
    return jnp.broadcast_to(result, self.batch_shape + self.event_shape)

  def stddev(self) -> Array:
    """Calculates the standard deviation (the square root of the variance)."""
    if isinstance(self.scale, diag_linear.DiagLinear):
      result = jnp.abs(self.scale.diag)
    else:
      result = jnp.sqrt(self.variance())
    return jnp.broadcast_to(result, self.batch_shape + self.event_shape)


MultivariateNormalLike = Union[
    MultivariateNormalFromBijector, tfd.MultivariateNormalLinearOperator]


def _squared_frobenius_norm(x: Array) -> Array:
  """Computes the squared Frobenius norm of a matrix."""
  return jnp.sum(jnp.square(x), axis=[-2, -1])


def _log_abs_determinant(d: MultivariateNormalLike) -> Array:
  """Obtains `log|det(A)|`."""
  if isinstance(d, MultivariateNormalFromBijector):
    log_det_scale = d.scale.forward_log_det_jacobian(
        jnp.zeros(d.event_shape, dtype=d.dtype))
  elif isinstance(d, tfd.MultivariateNormalLinearOperator):
    log_det_scale = d.scale.log_abs_determinant()
  else:
    raise ValueError(f'Unsupported distribution: {type(d)}.')
  return log_det_scale


def _inv_scale_operator(d: MultivariateNormalLike) -> Callable[[Array], Array]:
  """Gets the operator that performs `A^-1 * x`."""
  if isinstance(d, MultivariateNormalFromBijector):
    inverse_fn = jax.vmap(d.scale.inverse, in_axes=-1, out_axes=-1)
  elif isinstance(d, tfd.MultivariateNormalLinearOperator):
    inverse_fn = d.scale.solve
  else:
    raise ValueError(f'Unsupported distribution: {type(d)}.')
  return inverse_fn


def _scale_matrix(d: MultivariateNormalLike) -> Array:
  """Gets the full scale matrix `A`."""
  if isinstance(d, MultivariateNormalFromBijector):
    matrix = d.scale.matrix
  elif isinstance(d, tfd.MultivariateNormalLinearOperator):
    matrix = d.scale.to_dense()
  else:
    raise ValueError(f'Unsupported distribution: {type(d)}.')
  return matrix


def _has_diagonal_scale(d: MultivariateNormalLike) -> bool:
  """Determines if the scale matrix `A` is diagonal."""
  if (isinstance(d, MultivariateNormalFromBijector)
      and isinstance(d.scale, diag_linear.DiagLinear)):
    return True
  elif (isinstance(d, tfd.MultivariateNormalDiag) or
        (isinstance(d, tfd.MultivariateNormalFullCovariance) and
         d.parameters['covariance_matrix'] is None) or
        (isinstance(d, tfd.MultivariateNormalTriL) and
         not isinstance(d, tfd.MultivariateNormalFullCovariance) and
         d.parameters['scale_tril'] is None) or
        (isinstance(d, tfd.MultivariateNormalDiagPlusLowRank) and
         d.parameters['scale_perturb_factor'] is None)):
    return True
  return False


def _kl_divergence_mvn_mvn(
    dist1: MultivariateNormalLike,
    dist2: MultivariateNormalLike,
    *unused_args, **unused_kwargs,
    ) -> Array:
  """Divergence KL(dist1 || dist2) between multivariate normal distributions.

  Args:
    dist1: A multivariate normal distribution.
    dist2: A multivariate normal distribution.

  Returns:
    Batchwise `KL(dist1 || dist2)`.
  """

  num_dims = tuple(dist1.event_shape)[-1]  # `tuple` needed for TFP distrib.
  if num_dims != tuple(dist2.event_shape)[-1]:
    raise ValueError(f'Both multivariate normal distributions must have the '
                     f'same `event_shape`, but they have {num_dims} and '
                     f'{tuple(dist2.event_shape)[-1]} dimensions.')

  # Calculation is based on:
  # https://github.com/tensorflow/probability/blob/v0.12.1/tensorflow_probability/python/distributions/mvn_linear_operator.py#L384
  # If C_1 = AA.T, C_2 = BB.T, then
  #   tr[inv(C_2) C_1] = ||inv(B) A||_F^2
  # where ||.||_F^2 is the squared Frobenius norm.
  diff_lob_abs_det = _log_abs_determinant(dist2) - _log_abs_determinant(dist1)
  if _has_diagonal_scale(dist1) and _has_diagonal_scale(dist2):
    # This avoids instantiating the full scale matrix when it is diagonal.
    b_inv_a = jnp.expand_dims(dist1.stddev() / dist2.stddev(), axis=-1)
  else:
    b_inv_a = _inv_scale_operator(dist2)(_scale_matrix(dist1))
  diff_mean_expanded = jnp.expand_dims(dist2.mean() - dist1.mean(), axis=-1)
  b_inv_diff_mean = _inv_scale_operator(dist2)(diff_mean_expanded)
  kl_divergence = (
      diff_lob_abs_det +
      0.5 * (-num_dims +
             _squared_frobenius_norm(b_inv_a) +
             _squared_frobenius_norm(b_inv_diff_mean)))
  return kl_divergence


# Register the KL functions with TFP.
tfd.RegisterKL(
    MultivariateNormalFromBijector, MultivariateNormalFromBijector)(
        _kl_divergence_mvn_mvn)
tfd.RegisterKL(
    MultivariateNormalFromBijector, tfd.MultivariateNormalLinearOperator)(
        _kl_divergence_mvn_mvn)
tfd.RegisterKL(
    tfd.MultivariateNormalLinearOperator, MultivariateNormalFromBijector)(
        _kl_divergence_mvn_mvn)
